// (c) Copyright 2014 Hewlett-Packard Development Company, L.P.

(function () {

	// TODO per script version backward comp. & per engine version backward comp.
	Query(function (rule) { 
			if (rule.type == "Backward Compatibility") 
			return rule; 
	}).addEventsListeners({
		"registry/rule/added": function () { 
			this.selector = Query(this.selector).addEventListener(this.targetEvent, this.handler); 
		},

		"registry/rule/stateChanged": function () {
			if (this.disabled) {
				this.selector.removeEventListener(this.targetEvent, this.handler);
			} else {
				this.selector.addEventListener(this.targetEvent, this.handler);
			}
		}
	});

	Registry.addRule([
	{ 
		description: "Making scripts generated using engine version 11 to be compatible with engine version 12",
		selector: null,
		targetEvent	: "script/on-modify-load",
		handler: function (e) {
			var xmlDoc = e.target;
			var scriptStep = find("xpath:descendant-or-self::step[@section='Script']", xmlDoc.documentElement)[0];
			if (!scriptStep) {
				scriptStep = xmlDoc.createElement("step");
				var attrs = { "type": "block", "stepId": "step:{00000000-0000-0000-0000-000000000099}", "action": "action", "section": "Script", "level": "43" };
				for (var attr in attrs) {
					if (attrs.hasOwnProperty(attr)) { 
						scriptStep.setAttribute(attr, attrs[attr]);
					}
				}
				var topStep = find("xpath:step", xmlDoc.documentElement)[0];
				var sections = ["Init", "Action", "End"], index = 0;
				while (topStep.firstChild) {
					if (topStep.firstChild.nodeType == 1 && index < sections.length) 
						topStep.firstChild.setAttribute("section", sections[index++]);
					scriptStep.appendChild(topStep.firstChild);
				}

				var handlersStep = find("xpath:step[@section='Handlers']", scriptStep)[0];
				if (!handlersStep) {
					handlersStep = xmlDoc.createElement("step");
					var attrs = { "type": "block", "stepId": "step:{00000000-0000-0000-0000-000000000999}", "action": "Handlers", "section": "Handlers", "level": "43" };
					for (var attr in attrs) {
						if (attrs.hasOwnProperty(attr)) { 
							handlersStep.setAttribute(attr, attrs[attr]);
						}
					}
					var endStep = find("xpath:step[@section='End']", scriptStep)[0];
					scriptStep.insertBefore(handlersStep, endStep.nextElementSibling);
				}
				topStep.appendChild(scriptStep);
			}

			var librariesStep = find("xpath:descendant-or-self::step[@section='Libraries']", xmlDoc.documentElement)[0];
			if (!librariesStep) {
				librariesStep = xmlDoc.createElement("step");
				var attrs = { "type": "block", "stepId": "step:{00000000-0000-0000-0000-000000009999}", "action": "block", "section": "Libraries", "level": "43" };
				for (var attr in attrs) {
					if (attrs.hasOwnProperty(attr)) {
						librariesStep.setAttribute(attr, attrs[attr]);
					}
				}
				topStep.appendChild(librariesStep);

				var idents = find("xpath://ident", xmlDoc);
				for (var i = 0; i < idents.length; ++i) {
					try {
						// old scripts' ident json is not well formated
						var identObj = window.eval('(' + idents[i].textContent + ')');
						idents[i].textContent = JSON.stringify(identObj);
					} catch (ex) { }
				}
			}
		}
	},

	{
		description: "Update steps' end event",
		selector: null,
		targetEvent	: "script/on-modify-load",
		handler: function (e) {
			var xmlDoc = e.target;
			
			if (Ambiance.getBoolVal("LOAD_MODE")) {
			  // notice the last parameter for addEventListener is true - we want this handler to run before the handler in WebStepProgress.js
				Query(function (step) { 
					if (!RTS("step/endEvent", step)) 
						return step; }
				).addEventListener("step/start", function (e) {
					this.endEvent = Step.PROGRESS_STEP_NETWORK_COMPLETED;
				}, true);
			}
		}
	},

	{
		description: "Update steps' step id",
		selector: null,
		targetEvent	: "script/on-modify-load library/on-modify-load",
		handler: function (e) {
			var xmlDoc = e.target;
			var steps = find("xpath://step", xmlDoc);
			for (var i = 0; i < steps.length; ++i) {
				if (!steps[i].getAttribute('stepId')) {
					steps[i].setAttribute('stepId', steps[i].getAttribute('id'));
					steps[i].setAttribute('id', ""); // = undefined;
				}
			}

			var tOs = find("xpath://testObject", xmlDoc);
			for (var i = 0; i < tOs.length; ++i) {
				if (!tOs[i].getAttribute('testObjId')) {
					tOs[i].setAttribute('testObjId', tOs[i].getAttribute('id'));
					tOs[i].setAttribute('id', ""); // = undefined;
				}
			}
		}
	},
	
	{
		description: "Align the arguments of steps with the new JavaScript flag format",
		selector: null,
		targetEvent	: "script/on-modify-load library/on-modify-load",
		handler: function (e) {
			var xmlDoc = e.target;
			var steps = find("xpath://step", xmlDoc);
			for (var i = 0; i < steps.length; ++i) {
				var argumentsStr = steps[i].getAttribute("arguments");
				
				if (argumentsStr) {
					var stepArgs = JSON.parse(argumentsStr);
					var newStepArgs = {};
					for (var arg in stepArgs) {
						if (stepArgs.hasOwnProperty(arg)) {
							if (stepArgs[arg].value !== undefined) {
								return; // Will return once we find the first new value, 
							}			// 	if one value is in the new format then all of them are
							var value = stepArgs[arg];
							if(typeof value === "string"){
								if (value.match(/^\"[a-z,A-z,\s]+\"$/g)){
									newStepArgs[arg] = { value : value.substr(1, value.length - 2) };
								} else {
									newStepArgs[arg] = { value : value , evalJavaScript : true }; // old arguments always need to be evaluated
								}
							} else if (typeof value === "number"){
								newStepArgs[arg] = { value : value };
							} else {
								newStepArgs[arg] = { value : value , evalJavaScript : true };
							}
						}
					}
					steps[i].setAttribute("arguments", JSON.stringify(newStepArgs));
				}
			}
		}
	},

	{	
		description : "Update Runlogic structure",
		selector : null,
		targetEvent : "script/on-modify-load",
		handler : function (e) {
			var xmlDoc = e.target;
			var actionSteps = find("xpath:descendant-or-self::step[@action='action' and @section!='Script']", xmlDoc.documentElement);
			if (actionSteps.length != 0) {
				for (var i=0; i < actionSteps.length; ++i) {
					actionSteps[i].setAttribute("actionName", actionSteps[i].getAttribute("section"));
					actionSteps[i].removeAttribute("section");
				}
				var newRunLogicXml = ''.concat(
					'<step type="runLogic" stepId="step:{00000000-0000-0000-0000-000000000007}" action="default" level="43" name="default">',
						'<step type="block" stepId="step:{Init-Block000-0000-0000-000000000000}" overwriteUI="actionBlock" action="Init Block" level="1" index="1">',
							'<step type="control" stepId="step:{00000000-0000-0000-0000-000000000201}" action="Call Action" level="1" arguments="{&quot;Action Name&quot;:{&quot;value&quot;:&quot;Init&quot;}}" index="1"/>',
						'</step>',
						'<step type="block" stepId="step:{Run-Block0000-0000-0000-000000000000}" overwriteUI="actionBlock" action="Run Block" level="1" index="2">',
							'<step type="control" stepId="step:{00000000-0000-0000-0000-000000000202}" action="Call Action" level="1" arguments="{&quot;Action Name&quot;:{&quot;value&quot;:&quot;Action&quot;}}" index="1"/>',
						'</step>',
						'<step type="block" stepId="step:{End-Block0000-0000-0000-000000000000}" overwriteUI="actionBlock" action="End Block" level="1" index="3">',
							'<step type="control" stepId="step:{00000000-0000-0000-0000-000000000203}" action="Call Action" level="1" arguments="{&quot;Action Name&quot;:{&quot;value&quot;:&quot;End&quot;}}" index="1"/>',
						'</step>',
					'</step>');
				var newRunLogicXmlDoc = Utils.parseXml(newRunLogicXml);
				var newRunLogicStepOnNewXmlDoc = find("xpath:descendant-or-self::step[@type='runLogic']", newRunLogicXmlDoc.documentElement)[0];
				var newRunLogicStepOnOldXmlDoc = xmlDoc.importNode(newRunLogicStepOnNewXmlDoc, true);
				var oldRunLogicStep = find("xpath:descendant-or-self::step[@type='runLogic']", xmlDoc.documentElement)[0];
				oldRunLogicStep.parentNode.replaceChild(newRunLogicStepOnOldXmlDoc, oldRunLogicStep);
			}
		}
	}

], "Backward Compatibility");

})();