// (c) Copyright 2012 Hewlett-Packard Development Company, L.P.

(function () {
	"use strict";

	window.Step = function (typeOrObj, action, /*optional*/
		args, level /*=Step.LEVEL_NORMAL*/, identFlags
	) {
		// Note - This function returns an XML element with Step.prototype in its prototype chain.
		// Calling it with or without the 'new' keyword' are therefore equivalent
		var step = stepsDoc.createElement("step");
		var obj = window.TestObject && TestObject(typeOrObj);
		if (obj) {
			step.setAttribute("type", "testObject");
			step.setAttribute("testObject", obj.testObjId);
			step._obj = obj;
			obj.generateName();
		} else
			step.setAttribute("type", typeOrObj);

		step.setAttribute("stepId", "step:" + Utils.generateUUID());
		step.setAttribute("action", action);
		step.setAttribute("level", !isNaN(Number(level)) ? Number(level) : Step.LEVEL_NORMAL);
		
		step.identFlags = Object.extend({}, identFlags);
		if (identFlags) 
			step.setAttribute("identFlags", JSON.stringify(identFlags));
		
		step._args = {};
		for (var arg in args) {
			if (args.hasOwnProperty(arg)) {
				Step._setArgValue.call(step, arg, args[arg]);
			}
		}
		step.setAttribute("arguments", JSON.stringify(step._args));
		
		Step.init(step);
		
		Event.dispatch("step/created", step);
		return step;
	}

	Object.extend(Step, {
		LEVEL_NORMAL : 1,
		LEVEL_HIGH : 2,
		LEVEL_MAXIMUM : 3,
		LEVEL_HIDDEN : 42, // The Answer to the Ultimate Question of Life, the Universe, and Everything
		LEVEL_INTERNAL : 43, // (The Answer to the Ultimate Question of Life, the Universe, and Everything) + 1
		// never seen in the UI, always run
		RETCODE_SKIPPED : -2,
		RETCODE_ABORTED_SUCCEEDED : -1,
		RETCODE_SUCCEEDED : 0,
		RETCODE_ABORTED_FAILED : 1,
		RETCODE_FAILED : 2,

		// Generic extended info for errors
		EXT_RET_INFO_INVALID_ARGUMENTS : "Invalid Arguments",
		EXT_RET_INFO_INVALID_ARGUMENTS_RUNTIME : "Invalid Arguments Runtime",
		EXT_RET_INFO_NOT_IMPLEMENTED : "Not Implemented",
		EXT_RET_INFO_TIMED_OUT : "Timed Out",
		EXT_RET_INFO_PROGRESS_ERROR : "Progress Error",
		EXT_RET_INFO_NETWORK_TIMEOUT : "Network Timeout",

		// Generic extended info for non errors
		EXT_RET_INFO_STOPPING : "Stopping",
		EXT_RET_INFO_ERROR_IN_ANOTHER_STEP : "Error in Another Step",
		EXT_RET_INFO_ERROR_IN_HANDLER : "Error in Handler",

		STATE_IDLE : undefined,
		STATE_RUNNING : [],
		STATE_PENDING : [],
		STATE_WAITING : [],

		STATE_RUNNING_FROM : [],
		STATE_PAUSING : [],
		STATE_STOPPING : [],

		PROGRESS_STEP_STARTED : "Step started",
		PROGRESS_ACTION_STARTED : "Action started",
		PROGRESS_SCROLLED_INTO_VIEW : "Scrolled into view",
		PROGRESS_ACTION_ETA : "Action ETA",
		PROGRESS_ACTION_COMPLETED : "Action completed",
		PROGRESS_STEP_SYNC_NETWORK_COMPLETED : "Step synchronous network completed",
		PROGRESS_STEP_COMPLETED : "Step completed",
		PROGRESS_STEP_NETWORK_COMPLETED : "Step network completed",
		PROGRESS_DOM_CONTENT_LOADED : "DOM Content Loaded",
		PROGRESS_DOCUMENT_LOADED : "Document Loaded",
		PROGRESS_TIMERS_STARTED : "Timers Started",
		PROGRESS_TIMERS_ENDED : "Timers ended",
		PROGRESS_DIALOG_OPENED : "Dialog opened",
		PROGRESS_AFTER_STEP_ENDED : "After step ended",
		//new progress types for start operation notifications
		PROGRESS_WAIT_FOR_NETWORK : "Waiting for network",
		PROGRESS_WAIT_FOR_DOCUMENT : "Waiting for document load",
		PROGRESS_WAIT_FOR_DOM_CONTENT : "Waiting for DOM content load",
		PROGRESS_WAIT_FOR_DIALOG_OPEN : "Waiting for dialog open",
		PROGRESS_STEP_SYNC_NETWORK_ALMOST_COMPLETED : "Step synchronous network almost completed",
		PROGRESS_STEP_NETWORK_ALMOST_COMPLETED : "Step network almost completed",

		init : function (step) {
			step.state = Step.STATE_IDLE;
			if (step.__proto__)
				step.__proto__ = Step.prototype;
			else {
				step.oldSetAttribute = step.setAttribute;
				Object.extend(step, Step.prototype);
				step.origInsertBefore = step.insertBefore;
			}

			var stepExts = Step.getRulesByType(step.getAttribute("type"), step);

			for (var i = 0; i < stepExts.length; i++)
				Object.extend(step, stepExts[i]);

			Step.postExtendFix(step);

			// Attaching event listeners to the steps (and not the document), in order to support
			// notifications on disconnected steps (browser events).
			step.addEventListener("DOMNodeInserted", nodeInserted, false);
			step.addEventListener("DOMNodeRemoved", nodeRemoved, false);
			step.addEventListener("DOMAttrModified", attrModified, false);
		},

		clear : function () {
			var topStep = Step("block", "block", {
					comment : "TopStep"
				}, Step.LEVEL_NORMAL);
			stepsDoc.replaceChild(topStep, Step.steps);
			Step.steps = topStep;
			Step.sections = {};
		},

		getRulesByType : function(type, target) {
			var matchedRules = [];
			var rules = ruleByTypeTable[type];
			if (rules) {
				for (var j = 0; j < rules.length; j++) {
					var selector = rules[j].selector;
					if (selector.evaluate(target).indexOf(target) != -1) {
						matchedRules.push(rules[j].value);
					}
				}
			}
			return matchedRules;
		},

		adoptByStepRoot : function (rootStep) {
			if (!rootStep)
				return null;
			var mainStepsNode = stepsDoc.importNode(rootStep, true);

			// Remove all text nodes
			var textNodes = find("xpath:descendant::text()", mainStepsNode);
			for (var i = 0; i < textNodes.length; i++)
				textNodes[i].parentNode.removeChild(textNodes[i]);

			var steps = find("xpath:descendant-or-self::step", mainStepsNode);
			for (var i = 0; i < steps.length; i++) {
				var step = steps[i];

				var textNodes = find("xpath:child::text()", step);
				for (var j = 0; j < textNodes.length; j++)
					textNodes[j].parentNode.removeChild(textNodes[j]);

				Step.init(step);
				step._obj = TestObject(step.getAttribute("testObject"));
				step._args = JSON.parse(step.getAttribute("arguments")) || {};
				step.identFlags = JSON.parse(step.getAttribute("identFlags")) || {};
			}

			var sections = find("xpath:descendant-or-self::step[@section]", mainStepsNode);
			sections.forEach(function (sectionStep) {
				Step.sections[sectionStep.getAttribute("section")] = sectionStep;
				sectionStep.section = sectionStep.getAttribute("section");
			});

			return mainStepsNode;
		},

		import : function (rootStep, sectionName) {
			if (!rootStep)
				return;
			var mainStepsNode = Step.adoptByStepRoot(rootStep);
			if (!sectionName) {
				stepsDoc.replaceChild(mainStepsNode, Step.steps);
				Step.steps = mainStepsNode;
			} else {
				Step.sections[sectionName].appendChild(mainStepsNode);
			}

			return mainStepsNode;
		},

		export : function (step) {
			return Utils.nodeToXml(step);
		},

		getActionByName : function (name) {
			//TODO libs remove this method and don't use it for library
			var scriptAction = Step.steps.firstChild;
			var actionStep = find('xpath:step[@action="action" and @actionName="' + name + '"]', scriptAction)[0];
			if (actionStep)
				return actionStep;
			else
				return find('xpath:descendant-or-self::step[@type="library" and @action="' + name + '"]', Step.steps)[0];

		},
		
		getAllActionNames : function() {
			var allActionSteps =  find('xpath:descendant::step[@action="action"]', Step.steps);
			var retArr = [];
			for (var i=0; i < allActionSteps.length ; ++i) {
				var actionName = allActionSteps[i].getAttribute("actionName");
				if (actionName) {
					retArr.push(allActionSteps[i].getAttribute("actionName"));
				}
			}
			return retArr;
		},

		getStepsWithinSection : function(actionName, level) {
			actionName = actionName || "Action";
			level = (typeof level == "number" ? level : Step.LEVEL_NORMAL);
			var query = "xpath:descendant::step[@level <= " + level + "]";
			var stepsInLevel = find(query, this.getActionByName(actionName));
			return stepsInLevel;
		},

		getFunctionSteps : function (name) {
			return find("xpath:descendant::step[@type='function'".concat(name !== undefined ? (" and @action='" + name + "'") : "", "]"), stepsDoc);
		},

		toString : function () {
			return "The Step class"
		},

		onReadyStateUpdated : function (e) {
			var retVal = true;
			var step = e.target;
			if (step.readyState._aborted) {
				retVal = false;
				delete step.readyState._aborted;
			}
			else if (step.readyState.isReady) {
				if (step.type == "handler" && !step.preemptive) // Do not allow other handlers to interfere?
					disableHandlers = true;

				// Note - when calling the state setter the step is pushed into the Step.STATE_* array
				step.state = Step.STATE_RUNNING;
				step.run.callAsync(step);
			} else {
				var timeout = RTS("step/readyTimeout", step);
				if (timeout === undefined || (new Date()).getTime() - step.startTime <= timeout * 1000) {
					retVal =  false;
				}
				else {
					// Note - when calling the state setter the step is pushed into the Step.STATE_* array
					step.state = Step.STATE_RUNNING;

					if (Script.state == Step.STATE_STOPPING)
						step.end(Step.RETCODE_ABORTED_SUCCEEDED, {
							extRetInfo : Step.EXT_RET_INFO_STOPPING
						});
					else
						step.onReadyTimeout();
				}
			}
			Event.dispatch("step/processed", step, { retVal : retVal});
		},
		
		_setArgValue : function(name, value) { // Note - in fact a method and not a static function that is called for step instances 
			if (value !== undefined) {
					this._args[name] = this._args[name] || {};
					this._args[name].value = value;
				} else { 
					delete this._args[name];
				}
		}
	});

	Step.prototype = Property.initObj({
			stepId : Property({
				readonly : true
			}),
			type : Property({
				readonly : true
			}),
			action : Property(),
			level : Property({
				type : "number"
			}),
			endEvent : Property(),
			optional : Property({
				type : "boolean"
			}),
			disabled : Property({
				type : "boolean"
			}),
			breakpoint : Property({
				type : "boolean"
			}),
			duration : Property({
				type : "number",
				attrName : "manualDuration"
			}),
			useRecDuration : Property({
				type : "boolean"
			}),
			unfinalized : Property({
				type : "boolean"
			}),

			get obj() {
				return this._obj;
			},
			set obj(obj) {
				obj = TestObject(obj);
				obj.generateName();
				this._obj = obj;
				this.callOrigSetAttribute("testObject", obj.testObjId);
			},

			get query() {
				return (this._query || (( this._query = Query(this)) && this._query));
			},

			disconectQuery : function() {
				this._query = null;
			},

			get autoEndEvent() {
				return this.getAttribute('autoEndEvent');
			},
			set autoEndEvent(eve) {
				if (eve === null || eve === undefined) {
					this.removeAttribute('AllowXHROpenedRequest');
					this.removeAttribute('AllowSyncOpenedRequest');
					this.removeAttribute('autoEndEvent');
				} else
					this.setAttribute('autoEndEvent', eve);
			},

			get callingStep() {
				if (this.state == Step.STATE_IDLE)
					this._callingStep = undefined;
				return this._callingStep;
			},

			// Should be async for DRRE
			get evaledArgs() {
				var evaledArgs = {};
				for (var arg in this._args)
					if (this._args.hasOwnProperty(arg)) {
						this.evaluateArg(arg);
						evaledArgs[arg]  = this.getEvaledArg(arg);
					}
				return this._evaledArgs = evaledArgs;
			},


			get index() {
				return Number(this.getAttribute("index") || "0")
			}, // e.g. internal if group has no Index

			set index(index) {
				if (this.getAttribute("index") != index) {
					this.setAttribute("index", index);
				}
				Event.dispatch("step/indexed", this);
			},

			get fullIndex() {
				// TODO libs
				if (this == Step.steps || !this.parentNode || this.section)
					return "";

				var fullParentIndex = this.parentNode.fullIndex;
				var index = this.index;

				if (fullParentIndex)
					return index ? fullParentIndex.concat(".", index) : fullParentIndex;
				return index ? String(index) : "";
			},

			get realStep() {
				return this;
			},

			get recordSnapshotFileName() {
				return this.stepId.replace(":", "-").concat(".png");
			},

			getState : function () {
				return this._state;
			},

			setState : function (state) {
				if (state == this._state)
					return;

				if (this.state != Step.STATE_IDLE)
					this.state.splice(this.state.indexOf(this), 1);
				if (state != Step.STATE_IDLE)
					state.push(this);

				var prevState = this._state;
				this._state = state;
				Event.dispatch("step/stateChanged", this, {
					prevState : prevState
				});
			},

			getStateStr : function (state) {
				if (state == Step.STATE_IDLE)
					return "STATE_IDLE";
				if (state == Step.STATE_RUNNING)
					return "STATE_RUNNING";
				if (state == Step.STATE_PENDING)
					return "STATE_PENDING";
				if (state == Step.STATE_WAITING)
					return "STATE_WAITING";
				if (state == Step.STATE_RUNNING_FROM)
					return "STATE_RUNNING_FROM";
				if (state == Step.STATE_PAUSING)
					return "STATE_PAUSING";
				if (state == Step.STATE_STOPPING)
					return "STATE_STOPPING";
				return "no state"
			},

			get readyState() {
				return {
					isReady : true
				};
			},

			get nextStep() {
				return this.parentNode && this.parentNode.nextStep;
			},

			get progressTypes() {
				var progs = RuleEngine.getRules("stepProgress/dispatch", this);
				var retVal = [Step.PROGRESS_ACTION_COMPLETED];
				for (var i = 0; i < progs.length; i++)
					if (progs[i].public)
						retVal.push(progs[i].progressType);
				return retVal;
			},

			get containerSection() {
				return find("xpath:string(ancestor::step[@type='library']/@action)", this)[0] ||
				find("xpath:string((ancestor::step[@actionName])[last()]/@actionName)", this)[0];

			},

			updateReadyState : function () {
				Event.dispatch("step/ready-state-updated", this);
			},

			start : function (callingStep, forceStart, async) {
				if (this.state == Step.STATE_RUNNING) {
					Log.debug('[Step] ' + this.toString() + '" already running');
					return;
				}

				if (Script.state == Step.STATE_STOPPING)
					return this.end(Step.RETCODE_ABORTED_SUCCEEDED, {
						extRetInfo : Step.EXT_RET_INFO_STOPPING
					}, true);

				var $this = this;
				if (callingStep && callingStep.tagName != "step")
					callingStep = undefined;

				if (!forceStart) {
					if (this._shouldSkip())
						return this.end(Step.RETCODE_SKIPPED, null, true);

					if (RTS("step/shouldbreak", this)) {
						var breakpoint = RTS("step/breakpoint", this);
						if (breakpoint) {
							if (callingStep)
								this._callingStep = callingStep;
							// Note - when calling the state setter the step is pushed into the Step.STATE_* array
							this.state = Step.STATE_PENDING;
							Event.dispatch("step/pause", this, {
								reason : breakpoint
							});
							return;
						}
					}
				}

				if (!this.run)
					return this.end(Step.RETCODE_FAILED, {
						extRetInfo : Step.EXT_RET_INFO_NOT_IMPLEMENTED
					}, true);

				this._callingStep = callingStep || this._callingStep;
				this.startTime = (new Date()).getTime();

				// Note - when calling the state setter the step is pushed into the Step.STATE_* array
				this.state = Step.STATE_WAITING;

				if (!async)
					Step.STATE_WAITING.unshift(Step.STATE_WAITING.pop());

				this.notifyProgress(Step.PROGRESS_STEP_STARTED);
				var evt = Event.dispatch("step/start", this);
				enterDispatchLoop(evt);
			},

			end : function (retCode, props, forceEnd) {
				if (this._abortListener) {
					this.query.removeEventListener("step/abort", this._abortListener);
					this._abortListener = undefined;
				}

				if (this._timeoutTimer) {
					clearTimeout(this._timeoutTimer);
					this._timeoutTimer = undefined;
					if (this._timeout) {
						this._timeout = undefined;
						return this.end(Step.RETCODE_FAILED, {
							extRetInfo : Step.EXT_RET_INFO_TIMED_OUT
						});
					}
				}

				if (!forceEnd && this.state == Step.STATE_IDLE)
					return;

				// Note - when calling the state setter the step is pushed into the Step.STATE_* array
				this.state = Step.STATE_IDLE;

				(function () {
					if (retCode <= Step.RETCODE_SUCCEEDED)
						this.notifyProgress(Step.PROGRESS_STEP_COMPLETED);

					var evt = new Event("step/end", this, {
							retCode : retCode,
							propsObj : props,
							args : this._evaledArgs
						});
					Object.extendWithoutOverwrite(evt, props).dispatch(); // Check this properties and see that we override none
					this.notifyProgress(Step.PROGRESS_AFTER_STEP_ENDED, {retCode : retCode});
					if (this.state == Step.STATE_IDLE)
						this._callingStep = undefined;
					this._evaledArgs = undefined;
					this._nestingLevel = undefined;
					this.startTime = undefined;
					this._readyState = undefined;

					// For handlers - return them to wait step in case they need to run more then once
					// For all other steps - decrease the nesting level
					enterDispatchLoop(evt);
				}).callAsync(this);
			},

			notifyProgress : function (progressType, props) {
				function endStep() {
					var retCode = props && props.retCode > 0 && (RTS("RTS.TruClient.EndEventCauseFailure") || props.extRetInfo === Step.EXT_RET_INFO_NETWORK_TIMEOUT) ? props.retCode : Step.RETCODE_SUCCEEDED;
					if (props && props.retCode)
						props.retCode = retCode;
					var delay = RTS("step/shoulddelay", $this) ? 200 : 0;
					Event.dispatch("step/stepstartdelay", {}, {});
					setTimeout(function () {
						Event.dispatch("step/stependdelay", {}, {});
						$this.end(retCode, Object.extend(props, {
								progressType : progressType
							}));
					}, delay);
				}

				function startTimeoutTimer() {
					if ($this.state == Step.STATE_IDLE)
						return;
					$this._timeoutTimer = setTimeout(function () {
						Event.dispatch("step/timeout", $this);
						$this._timeoutTimer = setTimeout(function () {
								$this.end(Step.RETCODE_FAILED, {
									extRetInfo : Step.EXT_RET_INFO_TIMED_OUT
								});
							}, 10000);
						$this._timeout = true;
						$this.abort();
					}, stepTimeout * 1000);
				}

				var $this = this;
				if (progressType == Step.PROGRESS_ACTION_STARTED) {
					this.startTime = (new Date()).getTime();

					var stepTimeout = RTS("step/timeout", this);
					if (stepTimeout) {
						$this.query.addEventListenerUntil("step/assist/request", "step/end", function () {
							clearTimeout($this._timeoutTimer);
							$this._timeoutTimer = undefined;
							$this.query.addEventListenerUntil("step/assist/response", "step/end", startTimeoutTimer);
						});

						startTimeoutTimer();
					}
				} else if (progressType == Step.PROGRESS_ACTION_COMPLETED) {
					if (Script.state == Step.STATE_STOPPING)
						return this.end(Step.RETCODE_ABORTED_SUCCEEDED, {
							extRetInfo : Step.EXT_RET_INFO_STOPPING
						}, true);

					this._abortListener = function (e) {
						$this.end(Step.RETCODE_ABORTED_SUCCEEDED, {
							extRetInfo : Step.EXT_RET_INFO_STOPPING
						});
					};
					this.query.addEventListener("step/abort", this._abortListener);
					if (progressType != RTS("step/endEvent", this))
						this.notifyWaitForEndEvent(RTS("step/endEvent", this));
				}

				Object.extend(new Event("step/progress", this, {
					progressType : progressType
				}), props).dispatch();

				var endEvent = RTS("step/endEvent", this);
				if (((endEvent == Step.PROGRESS_STEP_SYNC_NETWORK_COMPLETED &&
							progressType == Step.PROGRESS_STEP_SYNC_NETWORK_ALMOST_COMPLETED &&
							this.getAttribute("AllowSyncOpenedRequest") >= props.syncReqLen)) ||
					((endEvent == Step.PROGRESS_STEP_NETWORK_COMPLETED &&
							progressType == Step.PROGRESS_STEP_NETWORK_ALMOST_COMPLETED &&
							this.getAttribute("AllowXHROpenedRequest") >= props.xhrReqLen))) {
					endStep();
				}

				if (progressType == endEvent ||
					//For IE: Handle navigation errors when step's end event is "sync network completed"
					(progressType == Step.PROGRESS_STEP_NETWORK_COMPLETED && endEvent == Step.PROGRESS_STEP_SYNC_NETWORK_COMPLETED)) {
					endStep();
				}
			},

			notifyWaitForEndEvent : function (endEvent, props) {
				switch (endEvent) {
				case Step.PROGRESS_STEP_NETWORK_COMPLETED:
				case Step.PROGRESS_STEP_SYNC_NETWORK_COMPLETED:
					Event.dispatch("step/WaitEndEvent", this, {
						progressType : Step.PROGRESS_WAIT_FOR_NETWORK
					});
					break;
				case Step.PROGRESS_DOM_CONTENT_LOADED:
					Event.dispatch("step/WaitEndEvent", this, {
						progressType : Step.PROGRESS_WAIT_FOR_DOCUMENT
					});
					break;
				case Step.PROGRESS_DOCUMENT_LOADED:
					Event.dispatch("step/WaitEndEvent", this, {
						progressType : Step.PROGRESS_WAIT_FOR_DOM_CONTENT
					});
					break;
				case Step.PROGRESS_DIALOG_OPENED:
					Event.dispatch("step/WaitEndEvent", this, {
						progressType : Step.PROGRESS_WAIT_FOR_DIALOG_OPEN
					});
					break;
				default:
					break;
				}
				return;
			},

			abort : function () {
				if (this.state == Step.STATE_WAITING) {
					if (this.type == "handler") {
						// Note - when calling the state setter the step is pushed into the Step.STATE_* array
						this.state = Step.STATE_IDLE;
					} else
						this.end(Step.RETCODE_ABORTED_SUCCEEDED, {
							extRetInfo : Step.EXT_RET_INFO_STOPPING
						});
					dispatchIndex = 0;
				}
				Event.dispatch("step/abort", this);
			},

			dispatchEvaluateArgs : function (names, args, argsSchema) {
				var step = this;
				function onArgumentsEvaluated(argsResults){
					step._evaledArgs = step._evaledArgs || {};
					for (var i = 0; i < names.length; i++) {
						step._evaledArgs[names[i]] = argsResults[names[i]];
					}
					Event.dispatch("step/evalParams-finished", step)
				}
				
				if(!args) {
					args = {};
					for (var i = 0; i < names.length; i++) {
						args[names[i]] = step._args[names[i]];
					}
				}

				Event.dispatch("agent/evaluate-arguments", window, {
					stepId: step.stepId,
					action: step.action,
					stepArgs: args,
					stepArgsTypes: argsSchema || step.argsSchema,
					stepTestObj: step.obj,
					funcArgs: step.getFuncInArgs(),
					callback: onArgumentsEvaluated
				});
			},

			evaluateArg : function (name, args, argsSchema) {
				return this.dispatchEvaluateArgs([name], args, argsSchema);
			},
			
			setArg : function (name, value) {
				Step._setArgValue.call(this, name, value); 	
				this.setAttribute("arguments", JSON.stringify(this._args));
			},
			
			setIsArgJavaScriptCode : function(name, evalJavaScript) {
				this._args[name] = this._args[name] || {};
				this._args[name].evalJavaScript = evalJavaScript;
				this.setAttribute("arguments", JSON.stringify(this._args));
			},
			
			getEvaledArg : function (name) {
				return this._evaledArgs ? this._evaledArgs[name] : undefined;
			},

			getUnevaledArg : function(name) {
				var nameObj = this._args[name];
				return nameObj ? nameObj.value : undefined;
			},
			
			getIsArgJavaScriptCode : function(name) {
				var nameObj = this._args[name];
				if (!nameObj)
					return (this.argsSchema && this.argsSchema[name] && this.argsSchema[name].javaScriptOnly) ? true : false;
				return nameObj.evalJavaScript || this.argsSchema[name].javaScriptOnly ? true : false;
			},
			
			getActiveArg : function(name) {
				var retValue = this.getEvaledArg(name);
				
				if (retValue instanceof Error)
					return "[Exception]";
					
				if (retValue !== undefined)
					return retValue;
				
				return this.getUnevaledArg(name);
			},
			
			getAllAvailableArgs : function() {
				var retArr = [];
				for (var argName in this._args) {
					if (this._args.hasOwnProperty(argName)) {
						retArr.push(argName);
					}
				}
				return retArr;
			},
			
			copyArgsFrom : function(otherStep) {
				this._args = {};
				Object.extend(this._args, otherStep._args);
			},

			clone : function () {
				var step = this.cloneNode(true);
				var oldSteps = find("xpath:descendant-or-self::step", this);
				var newSteps = find("xpath:descendant-or-self::step", step);

				for (var i = 0; i < newSteps.length; i++) {
					var curStep = newSteps[i];
					curStep.setAttribute("stepId", "step:" + Utils.generateUUID());
					Step.init(curStep);
					Object.extend(curStep, oldSteps[i]);

					// Do not let the new step.query object point to the old step
					curStep.disconectQuery();

					// for the IE xpath third party , infact not needed in FF
					curStep.__jsxpath_id__ = undefined;
					curStep._args = {};
					for(var arg in oldSteps[i]._args) {
						Object.extend(curStep._args[arg] = {}, oldSteps[i]._args[arg]);
					}
					Object.extend(curStep.identFlags = {}, oldSteps[i].identFlags);
				}

				Event.dispatch("step/cloned", step);
				return step;
			},

			addAlternative : function (step, asDefault) {
				if (this.parentNode && this.parentNode.type == "alternative")
					return this.parentNode.addAlternative(step, asDefault);

				// 'Alternative' StepExt overrides this function so from now on we can assume this is a regular step
				if (this.isEqual(step))
					return this;
				var altStep = Step("alternative", null, null, this.level);
				altStep.setAttribute("activeStep", "0");
				if (this.parentNode) {
					var placeholderStep = Step("placeholder", null, null, Step.LEVEL_HIDDEN);
					this.parentNode.replaceChild(placeholderStep, this);
				}
				altStep.addAlternative(this);
				if (placeholderStep)
					placeholderStep.parentNode.replaceChild(altStep, placeholderStep);

				return altStep.addAlternative(step, asDefault);
			},

			delegateRun : function (step) {
				function delegateProgress(e) {
					if (e.progressType != Step.PROGRESS_STEP_STARTED && e.progressType != Step.PROGRESS_STEP_COMPLETED)
						e.clone($this).dispatch();
				}

				var $this = this;
				step.query.addEventListener("step/progress", delegateProgress);
				step.query.addOnetimeEventListener("step/end", function (e) {
					step.query.removeEventListener("step/progress", delegateProgress);
					$this.end(e.retCode, e.propsObj);
				});

				step.start(this);
			},

			setAttribute : function (name, value) {
				if (name == "testObject")
					return this.obj = value;
				this.applyOrigSetAttribute(arguments);
			},

			validateArgs : function (args, argsSchema) {
				var errors = Schema.validate(args, argsSchema || this.argsSchema);
				if (!errors.length)
					return true;
				this.end(Step.RETCODE_FAILED, {
					extRetInfo : Step.EXT_RET_INFO_INVALID_ARGUMENTS,
					arguments : errors
				});
				return false;
			},

			isEqual : function (step) {
				if (step.tagName != "step" || this.type != step.type)
					return false;
				if (this == step)
					return true;
				return this._isEqual && this._isEqual(step);
			},

			getRealStep : function () {
				return this.realStep;
			},

			getCatchStep : function (errCode) {
				function checkErrorType(catchType, errType) {
					if (catchType == "Any")
						return true;
					switch (errType) {
					case "Object Not Found":
					case "Multiple Objects Found":
						return catchType == "Object Identification";

					case Step.EXT_RET_INFO_INVALID_ARGUMENTS:
						return catchType == "Step Arguments";

					case Step.EXT_RET_INFO_TIMED_OUT:
						return catchType == "Step Timeout";

					default:
						return catchType == "Step Action";
					}
				}

				function processNextStep(nextStep) {
					function onParamsEvaled() {
						var evaledArgs = step._evaledArgs;
						Schema.validate(evaledArgs, step.argsSchema);
						if (checkErrorType(evaledArgs["Error Type"], errCode))
							return Event.dispatch("step/catch-step", step, { errorStep: step});
						return processNextStep(step);
					}

					do {
						step = step.nextElementSibling;
					} while (step && step._shouldSkip());

					if (step && step.type == 'block' && step.getAttribute("catch")) {
						//for default Error type ("Any")
						if (step._args["Error Type"] == undefined) {
							step._evaledArgs = {};
							onParamsEvaled();
						} else {
							step.query.addOnetimeEventListener("step/evalParams-finished", onParamsEvaled);
							step.evaluateArg("Error Type");//Async in DRRE
						}
					} else {
						return Event.dispatch("step/catch-step", step, { errorStep: undefined});
					}
				}

				var step = this.parentNode && this.parentNode.type == "alternative" ? this.parentNode : this;
				if (this.type == "block" && this.getAttribute("catch"))
					return Event.dispatch("step/catch-step", step, { errorStep: undefined});

				processNextStep(step);
			},

			getParentFunction : function() {
				var step = this;
				while (step) {
					if (step.type == "function")
						return step;
					step = step.parentNode;
				}
				return undefined;
			},

			getFuncInArgs : function () {
				var funcStep = this.getParentFunction();
				return funcStep ? funcStep.funcArgs : undefined;
			},

			onReadyTimeout : function () {
				return this.end(Step.RETCODE_FAILED);
			},

			onNetworkTimeout : function (xhrReqs, Reqs) {
				Event.dispatch("step/network/timeout", this, {
					xhrReqs : xhrReqs,
					requests : Reqs
				});
			},

			toString : function () {
				return (this.fullIndex).concat(": ", this.toMarkup().replace(/<[^>]*>/g, "").replace(/\n|\t/g, " ").replace(/\u25CF/g, "*").xmlDecode());
			},

			toMarkup : function () {
				if (!this._toMarkup)
					return "[Step]";
				var str = this._toMarkup();
				str = str.replace(/<arg name='([^']*)'><\/arg>/g, "<arg missing='true' name='$1'>[$1]<\/arg>");
				str = str.replace(/<arg name='([^']*)'>([^<]{43,})<\/arg>/g, function (str, p1, p2) {
						return "<arg name='".concat(p1, "'>", p2.substr(0, 24), '...', p2.substr(p2.length - 16), "<\/arg>");
					});
				str = str.replace(/<testObj>([^<]{43,})<\/testObj>/g, function (str, p1) {
						return "<testObj>".concat(p1.substr(0, 24), '...', p1.substr(p1.length - 16), "<\/testObj>");
					});
				return str;
			},

			setTooltipValue : function (element, value) {
				if (typeof value == "string" && value) {
					var btValue = value.replace(/([",\\])/gm, function (str, p1) {
							return "\\" + p1;
						}).replace(/\n/gm, " ");
					element.bt({
						contentSelector : "\"" + btValue + "\"",
						trigger: ['none']
					});
				}
			},

			_shouldSkip : function () {
				return this.disabled || (this.level != Step.LEVEL_INTERNAL && (this.level > Step.steps.level ||
						(Script.state == Step.STATE_RUNNING_FROM && this.state != Step.STATE_RUNNING_FROM && this._shouldSkipFromStep())));
			},

			_shouldSkipFromStep : function () {
				return true;
			},

			randMinTime : function (duration) {
				duration = duration && Number(duration);
				if (isNaN(duration) || duration <= 0) {
					return 0;
				}
				if (RTS("RTS.TruClient.MinTimeEnableRandom")) {
					var low = Number(RTS("RTS.TruClient.MinTimeRandomLow"));
					var high = Number(RTS("RTS.TruClient.MinTimeRandomHigh"));
					var rand_percentage = Math.min(low, high) + Math.floor(Math.random() * (Math.abs(high - low) + 1)); //rand in range [low,high]
					duration = (duration * (rand_percentage / 100)).toFixed(3) - 0;

					if (RTS("RTS.TruClient.MinTimeEnableRandomLimit")) {
						var limit = Number(RTS("RTS.TruClient.MinTimeRandomLimitValue"));
						if (duration > limit)
							duration = limit;
					}
					Event.dispatch("step/mintime-rand", this, {
						duration : duration
					});
				} else {
					Event.dispatch("step/mintime", this, {
						duration : duration
					});
				}
				return duration;
			}

			/*	shouldBreakAtBP : function(){
			return (this.level != Step.LEVEL_INTERNAL && this.type != 'function') ||
			(this.level == Step.LEVEL_INTERNAL && this.type == 'testObject' && this.parentNode.type != 'alternative');
			},

			shouldDelayOnEnd : function(){
			return this.shouldBreakAtBP() && this.type != 'callFunction' && this.type != 'block' && this.type != 'If2');
			}*/
		});

	Query.statics.step = function (query) {
		return Query.statics.xpath("//step[@stepId='step:" + query + "']", Step.steps);
	};

	Query(function (rule) {
		if (rule.type == "step/extend")
			return rule;
	}).addEventListener(
		"registry/rule/added", function (e) {
		var props = Object.extend({}, this);
		delete props.selector;
		delete props.stepType;
		delete props.type;
		for (var i = 0; i < this.stepType.length; ++i) {
			var rules = ruleByTypeTable[this.stepType[i]];
			if (!rules) rules = ruleByTypeTable[this.stepType[i]] = [];
			var rule = { selector : Query(this.selector), value: props };
			rules.push(rule);
		}
	});

	function onStepGen(e) {
		if (Script.recordingContainer) {
			this.notifyExec(e.target);
			var step = this.handler(e);
			if (step && step.tagName == "step")
				Event.dispatch("step/generated", step);
		}
	}

	Query(function (rule) {
		if (rule.type == "Step Generators")
			return rule;
	}).addEventsListeners({
		"registry/rule/added" : function () {
			this._handler = onStepGen.bind(this);
			this.selector = Query(this.selector).addEventListener(this.eventTypes, this._handler);
		},

		"registry/rule/stateChanged" : function () {
			if (this.disabled)
				this.selector.removeEventListener(this.eventTypes, this._handler);
			else
				this.selector.addEventListener(this.eventTypes, this._handler);
		}
	});

	// Event listeners for step mutation
	function nodeInserted(e) {
		e.stopPropagation();
		e.cancelBubble = true;
		Event.dispatch("step/nodeInserted", e.target, {
			origEvent : e
		});
	}

	function nodeRemoved(e) {
		e.stopPropagation();
		e.cancelBubble = true;
		Event.dispatch("step/nodeRemoved", e.target, {
			origEvent : e
		});
	}

	function attrModified(e) {
		e.stopPropagation();
		e.cancelBubble = true;
		Event.dispatch("step/attrModified", e.target, {
			origEvent : e
		});
	}

	// handle steps indexing
	var fakeStep = {};

	function recalcChildIndexes() {
		var steps = find("xpath://step[@reindex]", stepsDoc);
		for (var i = 0; i < steps.length; i++) {
			steps[i].calcChildIndexs(steps[i].lastInsertedNode != fakeStep ? steps[i].lastInsertedNode : undefined);
			steps[i].removeAttribute("reindex");
			steps[i].lastInsertedNode = undefined;
		}
	}

	Query.addToBatch("step/nodeInserted step/nodeRemoved", "step/reindex");
	Query(function (batch) {
		if (batch.name == "step/reindex")
			return batch;
	})
	.addEventListener("batch/end", recalcChildIndexes);

	(function childStepAdded() {
		if (disableReindex) // do not try to reindex during script loading
			return;
		if (this.parentNode && this.parentNode.calcChildIndexs) {
			this.parentNode.setAttribute("reindex", '1');
			this.parentNode.lastInsertedNode = (!this.parentNode.lastInsertedNode) ? this : fakeStep;
		}
	}).bindToEvent("step/nodeInserted");

	(function childStepRemoved() {
		if (disableReindex) // do not try to reindex during script loading
			return;
		if (this.parentNode && this.parentNode.calcChildIndexs) {
			if (this.parentNode.lastChild == this)
				return;
			this.parentNode.setAttribute("reindex", '1');
			this.parentNode.lastInsertedNode = fakeStep;
		}
	}).bindToEvent("step/nodeRemoved");
	// end of handle steps indexing


	Query.addEventListener("step/nodeRemoved", function(e) {
		// Prevent circular reference of the step with his query
		if (e.target.disconectQuery)
			e.target.disconectQuery();
	});

	(function () {
		disableReindex++;
	}).bindToEvent("script/preload script/exportStart");

	(function () {
		disableReindex--;
	}).bindToEvent("script/postload script/exportEnd");

	var disableReindex = 0;

	(function addRecDuration() {
		function onStepCreated() {
			this.recTime = (new Date()).getTime() - startRecTime;
		}

		// TODO (amit): add support for alternative
		function calcBlock(block) {
			var steps = find("xpath:child::step[@level <= 3]", block),
			nextStep;
			for (var i = 0; i < steps.length; i++) {
				var step = nextStep = steps[i];
				if (step.type == 'block')
					calcBlock(step);
				else {
					do {
						var sibling = nextStep.nextElementSibling;
						nextStep = sibling ? (sibling.type == 'block' ? sibling.firstElementChild : sibling)
							 : (nextStep.parentNode != Script.recordingContainer ? nextStep.parentNode.nextElementSibling : null);
					} while (nextStep && (nextStep.level > step.level));

					if (nextStep) {
						var duration = Math.floor((nextStep.recTime - step.recTime) / 1000);
						if (duration >= 2)
							step.setAttribute("recDuration", duration);
					}
				}
				delete step.recTime;
			}
		}
		var startRecTime = (new Date()).getTime();

		var query = Query("xpath:self::*[@level <= 3]").addEventListener("step/created", onStepCreated);

		Query.addOnetimeEventListener("script/record/stop",
			function () {
			query.removeEventListener("step/created", onStepCreated);
			calcBlock(Script.recordingContainer);
		});

	}).bindToEvent("script/record/start");

	var stepsDoc = document.implementation.createDocument(null, null, null);

	// this flag is used in all browser types
	stepsDoc.isXml = true;

	// This code is not relevant to IE
	if (Step.prototype.__proto__)
		Step.prototype.__proto__ = stepsDoc.createElement("step").__proto__;

	Query.addOnetimeEventListener("step/moduleLoaded", function (e) {
		Step.steps = Step("block", "block", {
				comment : "TopStep"
			}, Step.LEVEL_NORMAL);
		stepsDoc.appendChild(Step.steps);
	});

	Query.addEventListener("loader/exposeRRE", function(e) {
		e.scope.Step = Step;
	});

	//The dispatch loop can run synchronically or not depending on the browser.
	function enterDispatchLoop(e) {
		function shouldProcessStep(step) {
			if (step._nestingLevel) { // if the step has a _nestingLevel then it is not a handler
				if (step._nestingLevel < nestingLevel)
					return false;
			} else if (disableHandlers)
				return false;
			return true;
		}

		function processStep(step) {
			Query.addOnetimeEventListener("step/ready-state-updated", Step.onReadyStateUpdated);
			step.updateReadyState();
		}

		function processNextStep (index) {
			if (index == Step.STATE_WAITING.length) {
				dispatchIndex = 0;
				asyncProcess();
				return;
			}
			var curStep = Step.STATE_WAITING[index];
			if (shouldProcessStep(curStep)) {
				curStep.query.addOnetimeEventListener("step/processed", function(e) {
					if (!e.retVal) {
						// Test object was not found yet (maybe later).
						// Lets run any handler currently in the loop (STATE_WAITING) and later try to find the object asynchronously.
						processNextStep(++index);
					}
				});
				// In Steps that have test object, in DRRE, we want to delegate the
				// step processing to the Agent side. In the StepExtTestObjectAsync we extend
				// the step with processStep function, in order to delegate it. In non DRRE we do not
				// add this function and we process the step as usual
				(curStep.processStep || processStep)(curStep);
			} else {
				processNextStep(++index);
			}
		}

		function asyncProcess() {
			var step = Step.STATE_WAITING[dispatchIndex];
			if (++dispatchIndex >= Step.STATE_WAITING.length)
				dispatchIndex = 0;
 			if (!step) {
				dispatchTimer = undefined;
			}
			else if (!shouldProcessStep(step)) {
				dispatchTimer = setTimeout(asyncProcess, 1000 / Step.STATE_WAITING.length);
 			}
 			else {
				step.query.addOnetimeEventListener("step/processed", function(e) {
					if (!e.retVal) {
						dispatchTimer = setTimeout(asyncProcess, 1000 / Step.STATE_WAITING.length);
					} else {
						dispatchTimer = undefined;
					}
				});
				// In Steps that have test object, in DRRE, we want to delegate the
				// step processing to the Agent side. In the StepExtTestObjectAsync we extend
				// the step with processStep function, in order to delegate it. In non DRRE we do not
				// add this function and we process the step as usual
				(step.processStep || processStep)(step);
			}
		}

		if (e.type == "step/start") {
			if (e.target.type != "handler") {
				e.target._nestingLevel = ++nestingLevel;
				processNextStep(0);
			}
		} else if (e.target.type != "handler") {
			--nestingLevel;
		} else { //step of type handler
			disableHandlers = false;

			if (Script.state == Step.STATE_RUNNING && !e.target.onetime && !e.target.disabled)
				e.target.start(null, false, true);

			if (e.retCode == Step.RETCODE_ABORTED_SUCCEEDED && e.extRetInfo == "Exit") {
				for (var i = 0; i < Step.STATE_WAITING.length; ++i) {
					if (Step.STATE_WAITING[i]._nestingLevel == nestingLevel) {
						Step.STATE_WAITING[i].end(Step.RETCODE_ABORTED_SUCCEEDED, {
							extRetInfo : e.extRetInfo
						});
						break;
					}
				}
			} else if (e.retCode <= Step.RETCODE_SUCCEEDED) {
				dispatchIndex = 0;
				asyncProcess();
			} else {
				for (var i = 0; i < Step.STATE_WAITING.length; ++i) {
					if (Step.STATE_WAITING[i]._nestingLevel == nestingLevel) {
						Step.STATE_WAITING[i].end(Step.RETCODE_ABORTED_FAILED, {
							extRetInfo : Step.EXT_RET_INFO_ERROR_IN_HANDLER
						});
						break;
					}
				}
			}
		}
	}

	var dispatchIndex = 0,
		nestingLevel = 0,
		dispatchTimer,
		disableHandlers = false;

	Query.addEventListener("script/replay/end", function() {
		if (dispatchTimer !== undefined)
			clearTimeout(dispatchTimer);
	});

	Query.addEventListener("agent/execute-step",
		function(args) {
			Step.executeStep(args/*stepId, action, stepArgs, stepTestObj, callback*/);
	});

	Query.addEventListener("agent/evaluate-test-object",
		function(args) {
			Step.evaluateTestObject(args/*identProfile, stepTestObj, callback*/)
	});

	Query.addEventListener("agent/evaluate-arguments",
		function(args) {
			Step.evaluateArgs(args/*stepId, stepTestObj, stepArgs, callback*/);
	});

	var ruleByTypeTable = {};

})();
