// (c) Copyright 2012 Hewlett-Packard Development Company, L.P.

(function() {
"use strict";

var transactionStates = {};
var STATE_OPEN = 0;
var STATE_OPEN_BY_API = 1;
var STATE_CLOSED = 2;
var STATE_CLOSED_BY_API = 3;
var STATE_UNDEFINED = 4;

window.Transaction = function(name, attrObj) {
	var trans = transDoc.createElement("transaction");
	if (name)
		trans.setAttribute("name", name);
	if (attrObj) {
		for (var name in attrObj) {
			trans.setAttribute(name, attrObj[name]);
		}
	}
	Transaction.init(trans);
	Transaction._transactions.appendChild(trans);
	return trans;
}

Object.extend(Transaction, {
	init : function(trans) {
		if (trans.__proto__)
			trans.__proto__ = Transaction.prototype;
		Object.extend(trans, Transaction.prototype); // Hack in FF , critical in IE
		trans.registerToDomEvents();

		// Give it a random ID.
		trans.transactionID = Math.floor((Math.random() * 1000)); //random number will be good for tracking.
		transactionStates[trans.name] = {};
		transactionStates[trans.name][trans.transactionID] = STATE_UNDEFINED;
	},

	clear : function() {
		// TODO libs enable to clear only transactions that related to specific steps
		var trash = find("xpath:transaction", Transaction._transactions);
		for (var i = 0; i < trash.length; i++)
			trash[i].clear();

		var trans = transDoc.createElement("transactions");
		transDoc.replaceChild(trans, Transaction._transactions);
		Transaction._transactions = trans;
		stepsLUT = {};
	},

	export : function(rootStep) {
		var tempNodes = Transaction._transactions.cloneNode(true);
		// TODO libs transactions without steps are removed in persistent but stay in memory
		var vertices = find("xpath:descendant::vertex", tempNodes);
		for (var i = 0; i < vertices.length; i++) {
			var vertexStep = find(vertices[i].getAttribute("step"))[0];
			if (vertexStep && (rootStep.compareDocumentPosition(vertexStep) & Node.DOCUMENT_POSITION_CONTAINED_BY)) continue;
			var trans = vertices[i].parentNode;
			trans.removeChild(vertices[i]);
			if (find("xpath:vertex", trans).length == 0)
				trans.parentNode.removeChild(trans);
		};

		return Utils.nodeToXml(tempNodes);
	},

	import : function(transactions) {
		//Transaction.clear();
		if (!transactions)
			return;
		transactions = find("xpath:transaction", transactions);
		for (var i = 0; i < transactions.length; i++) {
			var trans = transDoc.importNode(transactions[i], true);
			if (find("xpath:transaction[@name=\"" + trans.getAttribute("name").replace(/"/g, '&quot;') + "\"]", Transaction._transactions).length > 0) {
				Log.info("Transaction with the same name already exists");
				continue;
			}

			Transaction.init(trans);

			var vertexes = find("xpath:vertex", trans);
			for (var j = 0; j < vertexes.length; j++) {
				Vertex.init(vertexes[j]);
			}

			Transaction._transactions.appendChild(trans);
		}
	},

	merge : function(transactionsNodeOrArr) {
		if (!transactionsNodeOrArr) return;
		var transactions = (Array.isArray(transactionsNodeOrArr) ? transactionsNodeOrArr : find("xpath:transaction", transactionsNodeOrArr));
		for (var i = 0, transLen = transactions.length; i < transLen; i++) {
			var newTran = transactions[i];
			var currTran = find("xpath:transaction[@name=\"" + newTran.getAttribute("name").replace(/"/g, '&quot;') + "\"]", Transaction._transactions);
			currTran = (currTran.length > 0 ? currTran[0] : null);
			if (! currTran) {
				Transaction.init(newTran);
			}

			var newVertexes = find("xpath:vertex", newTran);

			for (var j = 0, vertexLen = newVertexes.length ; j < vertexLen; j++) {
				var newVertex = newVertexes[j];
				if (currTran && find("xpath:vertex[@type=\"" + newVertex.getAttribute("type").replace(/"/g, '&quot;') +
										// Add these 2 lines once we support multiple transaction start and end points in the UI
										//  "\" and @step=\"" + newVertex.getAttribute("step").replace(/"/g, '&quot;') +
										//  "\" and @progressType=\"" + newVertex.getAttribute("progressType").replace(/"/g, '&quot;') +
										"\"]", currTran).length > 0) {
					continue;
				}
				Vertex.init(newVertex);
				if (currTran) {
					currTran.appendChild(newVertex);
				}
			}
			if (! currTran) {
				Transaction._transactions.appendChild(newTran);
			}
		}
	},

	startFromAPI :function(name, type, flags) {
		var transToStart = 0;
		var trans = find("xpath:transaction", Transaction._transactions);
		for (var i = 0; i < trans.length; i++){
			if(name == trans[i].name){
				transToStart = trans[i];
				break;
			}
		}
		if(!transToStart){
			transToStart = new Transaction(name);
			transToStart.setAttribute("needToRemove", true); // Note - the getAttribute in IE is a string in any case !!
		}
		transToStart.type = type;
		transToStart.flags = flags;
		transToStart.start(true);
		return;
	},

	endFromAPI   : function(name, status) {
		var trans = find("xpath:transaction", Transaction._transactions);
		for (var i = 0; i < trans.length; i++){
			if(name == trans[i].name){
				trans[i].status = status;
				trans[i].end(status, true);
				return;
			}
		}
	},

	hasTransactionWithinSection : function(sectionName) {
		var transArr = find('xpath:transaction', Transaction._transactions);
		for (var i = 0; i < transArr.length; ++i) {
			var vertexes = transArr[i].vertexes;
			for (var j = 0; j < vertexes.length; ++j) {
				var step = vertexes[j].step;
				if (step.containerSection === sectionName) {
					return true;
				}
			}
		}
		return false;
	},

	removeTransIfNeeded : function(name){
		var trans = find("xpath:transaction", Transaction._transactions);
		for (var i = 0; i < trans.length; i++){
			if(name == trans[i].name){
				if(trans[i].hasAttribute("needToRemove")){
					trans[i].clear();
				}
				return;
			}
		}
	},

	removeEmptyTransactions : function() {
		var emptyTransactionsFound = false ;
		var transactions = find("xpath:transaction", Transaction._transactions);
		transactions.forEach(function(transaction) {
			if (transaction.childNodes.length === 0) {
				transaction.parentNode.removeChild(transaction);
				if(!emptyTransactionsFound) emptyTransactionsFound = true;
			}
		});
		return emptyTransactionsFound;
	},

	removeVertexes : function(step) {
		var transactions = find("xpath:transaction", Transaction._transactions);
		transactions.forEach(function(transaction) {
			if (transaction) {
				transaction.removeVertex("start", step);
				transaction.removeVertex("end", step);
			}
		});
	},

	getTransactionsByAttributes : function(attrObj) {
		var xpathAttrStr = "";
		for (var name in attrObj) {
			if (xpathAttrStr != "") {
				xpathAttrStr += " and ";
			}
			xpathAttrStr += "@" + name + "=\"" + attrObj[name] + "\"";
		}

		return find('xpath:transaction[' + xpathAttrStr + ']', Transaction._transactions);
	},

	registerAutoTransactions : function () {
		Query.addEventListenerUntil("TC.registerAutoTrans", "TC.endRunLogic", function (e) {
			function autoTransGen(name) {
				return function (e) {
					if (e.progressType == Step.PROGRESS_ACTION_STARTED) {
						// type currently 4, means automatic type; flags currently 2, means do not write to log
						Transaction.startFromAPI((typeof name == "function") ? name(e) : name, 4, 2);
						//flags currently 2, means do not write to log
					}
					if (e.progressType == Step.PROGRESS_STEP_COMPLETED) {
						Transaction.endFromAPI((typeof name == "function") ? name(e) : name, 0);
					}
				}
			}

			var mode = RTS("RTS.TruClient.AutomaticTransactions");

			if (!mode) return;
			
			var actionHandlers=[];
			var autoTransHandler;
			var actionID;
			var scriptAction = Step.steps.firstChild;
			var actionsRefArray = find("xpath:step[@action='action']", scriptAction);
			for (var i = 0; i < actionsRefArray.length; i++) {
				actionID = actionsRefArray[i].getAttribute("actionName");
				autoTransHandler = autoTransGen("Transaction_" + actionID);
				actionsRefArray[i].query.addEventListener("step/progress", autoTransHandler);
				actionHandlers.push(autoTransHandler);
			}


			if (mode == 2) {
				var stepHandler = autoTransGen(function (e) {
					return "Transaction_" + e.target.containerSection + "_Step_" + e.target.fullIndex;
				});
				var stepQuery = Query("xpath:self::*[@level <= 3]").addEventListener("step/progress", stepHandler);
			}

			Query.addOnetimeEventListener("script/replay/end", function () {
				for (var i = 0; i < actionsRefArray.length; i++) {
					actionsRefArray[i].query.removeEventListener("step/progress", actionHandlers[i]);
				}
				if (stepQuery) stepQuery.removeEventListener("step/progress", stepHandler);
			});
		});
	}
});

Transaction.listenersAdded = 0;
Transaction.transactionCounter = 0;


Transaction.prototype = {
	// Properties

	get name()				{ return this.getAttribute("name");	},
	set name(value)			{
		var transName = this.name;
		delete transactionStates[transName];
		this.setAttribute("name", value);
		transactionStates[this.name] = {};
		transactionStates[this.name][this.transactionID] = STATE_UNDEFINED;
	},

	get vertexes()			{ return find("xpath:vertex", this); },

	get endVertices()		{ return find("xpath:vertex[@type='end']", this); },
	get startVertices()     { return find("xpath:vertex[@type='start']", this); },

	// Public Methods
	clear : function() {
		var verticesToClear = find("xpath:vertex", this);
		for (var i = 0; i < verticesToClear.length; i++)
			verticesToClear[i].clear();

		if (this.parentNode) {
			this.parentNode.removeChild(this);

			if (transactionStates[this.name]) {
				delete transactionStates[this.name];
			}
		}
	},

	registerToDomEvents: function() {
		this.addEventListener("DOMNodeInserted", function(e) { new Event("transaction/nodeInserted", e.target, { origEvent : e }, Event.DISPATCH_SYNC);	}, false);
		this.addEventListener("DOMNodeRemoved", function(e) { new Event("transaction/nodeRemoved", e.target, { origEvent : e }, Event.DISPATCH_SYNC); }, false);
		this.addEventListener("DOMAttrModified", function(e) { new Event("transaction/attrModified" ,e.target, { origEvent : e }, Event.DISPATCH_SYNC);	}, false);
	},

	addVertex : function(typeOrVertex, step, progressType) {
		this.appendChild(typeOrVertex instanceof Vertex ? typeOrVertex : new Vertex(typeOrVertex, step, progressType));
	},

	removeVertex : function(typeOrVertex, step, progressType) {
		if (typeOrVertex instanceof Vertex) {
			if (typeOrVertex.parentNode == this) typeOrVertex.clear();
		}
		else {
			var vertexes = find('xpath:vertex[@type="'.concat(typeOrVertex, '" and @step="', step.stepId,
				progressType ? '" and @progressType="' + progressType : "",	'"]'), this);
			vertexes.forEach(function(vertex) {
				if (vertex) vertex.clear();
			});
		}
	},

	addStart	: function(step) {
		return this.addVertex("start", step, Step.PROGRESS_STEP_STARTED);
	},

	addEnd		: function(step) {
		return this.addVertex("end", step, Step.PROGRESS_AFTER_STEP_ENDED);
	},

	removeStart	: function(step, progressType) { return this.removeVertex("start", step, progressType); },
	removeEnd	: function(step, progressType) { return this.removeVertex("end", step, progressType); },


	onWaitStart : function(e) {
		var waitStartTime = e.time;
		Query.addOnetimeEventListener("step/wait/end step/abort", function (e) {
			if (e.time) { // when the wait step was aborted there is no time
				var thinkTime = e.time - waitStartTime;
				Event.dispatch("step/thinkTime", null, { thinkTime: thinkTime })
			}
		});
	},

	onPacing : function(e){
		var startTime = new Date().getTime();
		var pacingCause = e.cause;
		// The step/abort is for removing the listener, no wastedTime will be reported
		Query.addOnetimeEventListener("step/end/pending step/abort", function(e) {
			if(!Transaction.transactionCounter) {
				return;
			}
			var eventType, timeObj;
			var time = e.time - startTime;
			if (pacingCause == "pacing") {
				eventType = "step/wastedTime";
				timeObj = { wastedTime: time };
			}
			else if (pacingCause == "minTime") {
				eventType = "step/thinkTime";
				timeObj = { thinkTime: time };
			}
			else {
				return;
			}
			Event.dispatch(eventType, null, timeObj);
		},false);
	},

	onStepDelay : function (e){
		var startTime = new Date().getTime();
		Query.addOnetimeEventListener("step/stependdelay", function(e) {
			if(!Transaction.transactionCounter) {
				return;
			}
			var wastedTime = new Date().getTime() - startTime;
			Event.dispatch("step/wastedTime", null, { wastedTime: wastedTime });
		},false);
	},

	addStepsEventListners : function(){
		if(!Transaction.listenersAdded){
			Query.addEventListener("step/wait/start", this.onWaitStart,  false);
			Query.addEventListener("step/start/pending", this.onPacing, false);
			Query.addEventListener("step/stepstartdelay", this.onStepDelay, false);
		}
	 	Transaction.listenersAdded = true;
	},

	RemoveStepsEventListners : function(){
		if(Transaction.listenersAdded){
			Query.removeEventListener("step/wait/start", this.onWaitStart);
			Query.removeEventListener("step/start/pending", this.onPacing);
			Query.removeEventListener("step/stepstartdelay", this.onStepDelay);
		}
		Transaction.startWaitTime = 0;
		Transaction.listenersAdded = 0;
	},


		start : function(forceStart) {
			if (!forceStart && transactionStates[this.name][this.transactionID] == STATE_OPEN) {
					return;
			}
			
			if (!forceStart) {
				transactionStates[this.name][this.transactionID] = STATE_OPEN;
			} else {
				transactionStates[this.name][this.transactionID] = STATE_OPEN_BY_API;
			}
			
			Transaction.transactionCounter++;
			this.addStepsEventListners();
			this.startTime = new Date();
			new Event("TC.startTransaction", {}, { name: this.name }, Event.DISPATCH_SYNC);
			Event.dispatch("transaction/start", this);
	},

		end : function(transStatus, forceEnd) {
			if (!forceEnd && transactionStates[this.name][this.transactionID] == STATE_CLOSED) {
					return;
			}
        
			if (!forceEnd) {
				transactionStates[this.name][this.transactionID] = STATE_CLOSED;
			} else {
				transactionStates[this.name][this.transactionID] = STATE_CLOSED_BY_API;
			}
			
			this.endTime = new Date();
			new Event("TC.endTransaction", {}, { name:this.name, id: this.transactionID,
				status: (transStatus ? transStatus : this.status ), duration: this.endTime - this.startTime }, Event.DISPATCH_SYNC);
			this.type = undefined;
			Event.dispatch("transaction/end", this);
			if (Transaction.transactionCounter > 0) {
				Transaction.transactionCounter--;
			}
			if(Transaction.transactionCounter == 0) {
			  		this.RemoveStepsEventListners();
			}
			Transaction.removeTransIfNeeded(this.name);
			this.endTime = this.startTime = null;
	},

	isValid		: function() {
		var isEmpty = this.childNodes.length == 0;
		var properlyPaired = find('self::*[vertex[@type="start"] and vertex[@type="end"]]', this);

		return isEmpty || properlyPaired;
	}
}

function Vertex(type, step, progressType) {
	var vertex = transDoc.createElement("vertex");
	Vertex.init(vertex); //should be before settings vertex's attributes
	vertex.type = type;
	vertex.progressType = progressType;
	vertex.step = step;
	return vertex;
}

Vertex.init = function(vertex) {
	//if not IE
	if (vertex.__proto__ )
		vertex.__proto__ = Vertex.prototype;
	else
		Object.extend(vertex, Vertex.prototype);
	vertex.insertToStepsLUT();
	vertex.registerToDomEvents();
	vertex.bindInvoke();
}

Vertex.prototype = {
	// Properties
	get type()			{ return this.getAttribute("type");	},
	set type(value)			{ this.setAttribute("type", value);	},

	get step()			{ return find(this.getAttribute("step"))[0]; },
	set step(step)			{ this.setAttribute("step", step.stepId); },

	get progressType() 		{ return this.getAttribute("progressType"); },
	set progressType(value) 	{ this.setAttribute("progressType",value); },

	// Public Methods
	insertToStepsLUT: function() {
		var step = this.step;
		if (!step) return;
		if (!(step.stepId in stepsLUT))
			stepsLUT[step.stepId] = {};
		if (!(this.progressType in stepsLUT[step.stepId]))
			stepsLUT[step.stepId][this.progressType] = [];
		stepsLUT[step.stepId][this.progressType].push(this);
	},

	registerToDomEvents: function() {
		this.addEventListener("DOMNodeInserted", function(e) { new Event("transaction/vertex/nodeInserted", e.target, { origEvent : e }, Event.DISPATCH_SYNC);	}, false);
		this.addEventListener("DOMNodeRemoved", function(e) { new Event("transaction/vertex/nodeRemoved", e.target, { origEvent : e }, Event.DISPATCH_SYNC); }, false);
		this.addEventListener("DOMAttrModified", function(e) { new Event("transaction/vertex/attrModified" ,e.target, { origEvent : e }, Event.DISPATCH_SYNC);	}, false);
	},

	clear : function() {
		if (this.parentNode) this.parentNode.removeChild(this);
		var step = this.step;
		if (step) {
			var verticesInLUT = stepsLUT[step.stepId][this.progressType];
			verticesInLUT.splice(verticesInLUT.indexOf(this),1);
		}
	},

	bindInvoke : function() {
		switch (this.type) {
			case "start"	: this.invoke = this.parentNode.start	; break;
			case "end"		: this.invoke = this.parentNode.end		; break;
			default			: this.invoke = this._nothing			; break;
		}
	},

	// Private Methods
	_nothing: function() {;}
}

//
// Transaction Module Initialization
//
var stepsLUT = {};
var transDoc = document.implementation.createDocument(null, null, null);

// this flag is used in all the browser types
transDoc.isXml = true;

Transaction._transactions = transDoc.appendChild(transDoc.createElement("transactions"));

// If not IE
if (Transaction.prototype.__proto__ ) {
	Transaction.prototype.__proto__ = transDoc.createElement("transaction").__proto__;
	Vertex.prototype.__proto__ = transDoc.createElement("vertex").__proto__;
}
//
// Transaction DOM Events Generators
//
// IE - the DOM NodeInserted is not working on other elements either then the own element
/*

transDoc.addEventListener("DOMNodeInserted", function(e) {
	if (e.target.tagName == "transaction" || e.target.tagName == "vertex")
		Event(e.target.tagName == "transaction" ? "transaction/nodeInserted" : "transaction/vertex/nodeInserted",
			e.target, { origEvent : e }).dispatch();
	}, false);

transDoc.addEventListener("DOMNodeRemoved", function(e) {
	if (e.target.tagName == "transaction" || e.target.tagName == "vertex")
		Event(e.target.tagName == "transaction" ? "transaction/nodeRemoved" : "transaction/vertex/nodeRemoved",
			e.target, { origEvent : e }).dispatch();
	}, false);

transDoc.addEventListener("DOMAttrModified", function(e) {
	if (e.target.tagName == "transaction" || e.target.tagName == "vertex")
		Event(e.target.tagName == "transaction" ? "transaction/attrModified" : "transaction/vertex/attrModified",
			e.target, { origEvent : e }).dispatch();
	}, false);
*/
//

// Transaction Related Rules
Query.addEventsListeners({
	// RULE: Invoke Vertices During Replay
	"step/progress" : function(e) {
		// Ignore the progress when the step failed
		if (typeof e.retCode != "undefined" && e.retCode > Step.RETCODE_SUCCEEDED)
			return;

		var LUT = stepsLUT[this.stepId];
		var verticesToInvoke = LUT && LUT[e.progressType];
		if (verticesToInvoke) {
			for (var i = 0; i < verticesToInvoke.length; i++) {
				if (!verticesToInvoke[i].parentNode.status) {
					verticesToInvoke[i].parentNode.status = 0;
				}
				verticesToInvoke[i].invoke.call(verticesToInvoke[i].parentNode);
			}
		}
			
	},

	// RULE: When a vertex is created, associate an invoke function
	"transaction/vertex/nodeInserted" : function(e) { e.target.bindInvoke(); },

	// RULE: rebuild stepsLUT
	"transaction/nodeInserted transaction/vertex/nodeInserted transaction/nodeRemoved transaction/vertex/nodeRemoved transaction/attrModified transaction/vertex/attrModified" : function(e) {
		stepsLUT = {};
		var vertexes = find("xpath://vertex", Transaction._transactions);
		for (var i = 0; i < vertexes.length; i++) vertexes[i].insertToStepsLUT();
	},

	"loader/exposeRRE" :  function(e) {
		e.scope.Transaction = Transaction;
	},

	"RRE/getTransactions" : function(e) {
		e.transactions = Transaction._transactions;
	},
    
	"TC.startIteration TC.startRunLogic" : function (e) {
	
	        // Go over transaction states and reset!
	        for (var transName in transactionStates) {
	            for (var transID in transactionStates[transName]) {
	                transactionStates[transName][transID] = STATE_UNDEFINED; // Reset current state
	            }
	        }
	}
});

})();
