// (c) Copyright 2012 Hewlett-Packard Development Company, L.P.

(function() {

"use strict";

window.TestObject = function(obj, flags) {
	// Is obj a TestObject instance?
	if (!obj || obj.tagName == "testObject")
		return obj;

	// If obj is a TestObject testObjId, look it up in the repository
	if (typeof obj == "string")
		return activeRepo[obj];

	// Is the obj the browser?
	if (obj == TestObject.getActiveBrowserObj() && activeRepo[TestObject.ID_BROWSER])
		return activeRepo[TestObject.ID_BROWSER];

	// At this point obj is an AUT DOM element
	flags = flags || TestObject.TO_DEFAULT;

	// Check if obj has a valid TestObject link attached to it that still matches the object
	// Note : bitmap operation
	if (flags & TestObject.TO_CHECK_LINK && obj[TestObject.nameOfTOPtr] && obj[TestObject.nameOfTOPtr].tagName == "testObject") {
		var testObj = obj[TestObject.nameOfTOPtr];

		if (testObj == activeRepo[testObj.testObjId] && testObj.ident.checkTestObj(obj)) {
			testObj.object = obj;
			return testObj;
		}
	}

	// Search the repository for a matching TestObject
	// Note : bitmap operation
	if (flags & TestObject.TO_SEARCH_REPOSITORY) {
		for (var testObj = activeRepo.firstChild; testObj; testObj = testObj.nextSibling) {
			if (testObj.ident.isMatch(obj)) {
				testObj.object = obj;
				return testObj;
			}
		}
	}

	// Note : bitmap operation
	if (!(flags & TestObject.TO_CREATE))
		return null;

	// Create a new TestObject
	var testObj = TestObject._toDoc.createElement("testObject");
	testObj.addEventListener("DOMAttrModified", TestObject.onTOAttrModified, false);

	Object.setObjProto(testObj, TestObject.prototype);
	testObj.setAttribute("testObjId", "testObj:" + Utils.generateUUID());
	testObj.object = obj;
	testObj.cmds = {};
	testObj.props = {};

	if (flags & TestObject.TO_EXTEND) {
		var toExts = RuleEngine.getRules("testObject/extend", obj, obj.tagName.toUpperCase());
		for (var i = 0; i < toExts.length; i++) {
			var toExt = toExts[i];
			if (!toExt.disabled) {
				toExt.notifyExec(obj);
				if (toExt.role)
					testObj.addRole(typeof toExt.role == "function" ? toExt.role(obj) : toExt.role);
				if (toExt.methods)
					Object.extend(testObj, toExt.methods);
			}
		}
	}

	if (flags & TestObject.TO_CREATE_IDENT) {
		testObj.appendChild(TestObject._toDoc.createElement("idents"));
		testObj.ident = Ident(testObj, "Electors");
		Ident(testObj, "XPath");
		Ident(testObj, "Descriptors");
	}

	if (flags & TestObject.TO_ADD_TO_REPOSITORY) {
		activeRepo.appendChild(testObj);
		// Using activeRepo as a HashTable for testObjId -> testObj lookups (Not using document.getElementById() as it is
		// not supported in Firefox on XML document (bug #275196))
		activeRepo[testObj.testObjId] = testObj;

		//wrappedJSObject exist only in FF
		//if (TestObject.isWrappedJSObject(obj))
			obj[TestObject.nameOfTOPtr] = testObj;
	}

	return testObj;
}

Object.extend(TestObject, {
	TO_DEFAULT							: 0xFFFF,
	TO_CREATE							: 1 << 0,
	TO_CREATE_IDENT						: 1 << 1,
	TO_EXTEND							: 1 << 2,
	TO_CHECK_LINK						: 1 << 3,
	TO_SEARCH_REPOSITORY				: 1 << 4,
	TO_ADD_TO_REPOSITORY				: 1 << 5,

	ID_BROWSER							: "testObj:{00000000-0000-0000-0000-000000000001}",

	RETCODE_SUCCEEDED					: 0,
	RETCODE_FAILED						: 1,
	RETCODE_NOT_IMPLEMENTED				: 2,
	RETCODE_INVALID_ARGUMENTS			: 3,
	RETCODE_INVALID_ARGUMENTS_RUNTIME	: 4,
	RETCODE_ABORTED						: 5
});

TestObject.TO_NO_PERSISTENCE = (TestObject.TO_CREATE | TestObject.TO_EXTEND | TestObject.TO_CHECK_LINK | TestObject.TO_SEARCH_REPOSITORY);

Object.extend(TestObject, {
	clearAll : function() {
		TestObject._toDoc.replaceChild(TestObject._toDoc.createElement("testObjectsContainer"), TestObject._toDoc.firstChild);
		id2rep = {};
		TestObject.switchRepo(DEFAULT_REPOSITORY);
		Event.dispatch("repository/clear", TestObject);
	},

	clear : function() {
		while (activeRepo.hasChildNodes())
			activeRepo.removeChild(activeRepo.firstChild);
		Event.dispatch("repository/clear", TestObject);
	},

	import : function(testObjsElem, clear) {
		if (clear)
			TestObject.clear();
		var newTestObjsElems = find("xpath:testObject", testObjsElem);
		var newTestObjs = [];
		for (var i = 0; i < newTestObjsElems.length; i++) {
			var newTestObj = initTestObjectFromElem(newTestObjsElems[i]);
			if (newTestObj) newTestObjs.push(newTestObj);
		}

		// Creating identificaion in second pass as they can reference other test objects
		for (var i = 0; i < newTestObjs.length; ++i)
			initTestObjectIdents(newTestObjs[i]);
	},

	export : function(rootTOs) {
		if (!rootTOs) {
			var testObjs = activeRepo;
			for (var testObj = testObjs.firstChild; testObj; testObj = testObj.nextSibling)
				testObj.prepareToBeSerializedResistanceIsFutile();
		}
		else {
			var testObjs = TestObject._toDoc.createElement("testObjects");
			rootTOs.forEach(function(testObj) {
				testObj.prepareToBeSerializedResistanceIsFutile();
				testObjs.appendChild(testObj.cloneNode(true)); });
		}

		return serializer.serializeToString(testObjs);
	},

	getInventory : function() {
		var inventory = {};
		for (var testObj = activeRepo.firstChild; testObj; testObj = testObj.nextSibling) {
			var inventoryEntery = inventory[testObj.name] = (inventory[testObj.name] || []);
			inventoryEntery.push(
				{
					roles : testObj.roles,
					id : testObj.testObjId
				}
			);
		}

		return inventory;
	},

	gc : function(rootTOs) {
		// Remove test objects that are not part of the active repository
		var newRootTOs = [];
		for (var i = 0; i < rootTOs.length; ++i) {
			var rootTO = rootTOs[i];
			if (activeRepo[rootTO.testObjId]) {
				newRootTOs.push(rootTO);
				newRootTOs[rootTO.testObjId] = rootTO;
			}
		}

		// Add to newRootTOs test objects with a "live" object property
		for (var testObj = activeRepo.firstChild; testObj; testObj = testObj.nextSibling)
			if (testObj.object && !newRootTOs[testObj.testObjId]) {
				newRootTOs[testObj.testObjId] = testObj;
				newRootTOs.push(testObj);
			}

		var refferedTO = TestObject.collectReferredTOs(newRootTOs);

		// Sweep phase
		for (var testObj = activeRepo.firstChild; testObj; testObj = nextObj) {
			var nextObj = testObj.nextSibling;
			if (!refferedTO[testObj.testObjId]) {
				activeRepo.removeChild(testObj);
				delete activeRepo[testObj.testObjId];
			}
		}
	},

	collectReferredTOs : function (rootTOs) {
		// Add to rootTOs test objects referenced from other rootTOs
		for (var i = 0; i < rootTOs.length; i++) {	// Note - rootTOs array can change during the loop
			var testObj = rootTOs[i];
			testObj.prepareToBeSerializedResistanceIsFutile();
			var testObjIds = serializer.serializeToString(testObj).match(/testObj:{.{36}}/g);
			for (var j = 0; j < testObjIds.length; j++) {
				if (rootTOs[testObjIds[j]]) continue;
				var innerTestObj = TestObject(testObjIds[j]);
				rootTOs[innerTestObj.testObjId] = innerTestObj;
				rootTOs.push(innerTestObj);
			}
		}

		return rootTOs;
	},

	onTOAttrModified : function (e) {
		if (e.target.tagName == "testObject")
			Event.dispatch("testObject/attrModified", e.target, { origEvent : e });
	},

	cloneToRepo : function(testObjs, repoId, clear) {
		var origRepoId = TestObject.switchRepo(repoId);

		if (clear)
			TestObject.clear();

		var old2newID = {};
		testObjs.forEach(function(testObj) {
			var key = testObj.getAttribute("testObjId");
			var value = key == TestObject.ID_BROWSER ? key : "testObj:" + Utils.generateUUID();
			old2newID[key] = value });

		var newTestObjs = [];
		testObjs.forEach(function(testObj) {
			if (testObj.getAttribute("testObjId") == TestObject.ID_BROWSER) return;

			testObj.prepareToBeSerializedResistanceIsFutile();
			var toXMLStr = serializer.serializeToString(testObj).replace(/testObj:{.{36}}/g, function(str) { return old2newID[str]; });
			var toElem = find("xpath:testObject", Utils.parseXml(toXMLStr))[0];
			var newTestObj = initTestObjectFromElem(toElem);
			if (newTestObj) newTestObjs.push(newTestObj);
		});
		newTestObjs.forEach(initTestObjectIdents);
		TestObject.switchRepo(origRepoId);

		return old2newID;
	},

	switchRepo : function(repoId) {
		var oldTestObjs = activeRepo;
		repoId = repoId || DEFAULT_REPOSITORY;
		activeRepo = id2rep[repoId];
		if (!activeRepo) {
			var newTestObjs = TestObject._toDoc.createElement("testObjects");
			activeRepo = id2rep[repoId] = TestObject._toDoc.firstChild.appendChild(newTestObjs);
			activeRepo.repoId = repoId;
			TestObject.clear();
		}

		activeNameOfTOPtr = activeRepo.repoId.concat("-", loadingTiemstamp, "-", window.RREWindowName);
		return oldTestObjs && oldTestObjs.repoId;
	},

	removeRepo : function(repoId) {
		var repo = id2rep[repoId];
		if (repoId != DEFAULT_REPOSITORY && repo && repo != activeRepo) {
			repo.parentNode.removeChild(repo);
			delete id2rep[repoId];
		}
	},

	get nameOfTOPtr()	{ return activeNameOfTOPtr; },

	get activeRepo(){
		return activeRepo;
	},

	toString : function() { return "The TestObject class"},
	
	allTestObject : function() { return find("xpath:descendant-or-self::testObject", TestObject._toDoc.documentElement); }
});

TestObject.prototype = {
	get object() {
		TestObject.checkObjectReachability(this);

		// OK if this._object.children is undefined as long as we can ask the object about it
		return this._object;
	},

	set object(obj) {
		this._object = obj;
		if (obj && TestObject.isWrappedJSObject(obj)) obj[TestObject.nameOfTOPtr] = this;
	},

	get name() {
		return this.getAttribute("manualName") || this.getAttribute("autoName") || this.getAttribute("fallbackName") || "";
	},

	set name(value)			{ this.setAttribute("manualName", value); },

	get testObjId()				{ return this.getAttribute("testObjId"); },

	get icon()				{ return find("xpath:string(icon)", this)[0]; },

	get roles() {
		var retVal = [], roles = find("xpath:roles/role", this);
		for (var i = 0; i < roles.length; i++)
			retVal.push(roles[i].textContent);
		return retVal;
	},

	get refreshedObject() {
		var obj = this.object;	// object getter ensures that parent window of obj != null
		if (obj && TestObject.isObjExist(obj) &&
			!(obj.ownerDocument.compareDocumentPosition(obj) & Node.DOCUMENT_POSITION_DISCONNECTED) &&
				this.ident.checkTestObj(obj, Ident.PROFILE_EXACT)) return obj;

		var objs = this.evaluate();
		this.object = objs.length == 1 ? objs[0] : null;
		return this.object;
	},

	get ident()				{ return find("xpath:idents/ident[@type=../@active]", this)[0];	},
	set ident(val)			{ find("xpath:idents", this)[0].setAttribute("active", val.getAttribute("type")); },

	get rect() {
		return this._objRect;
	},

	set rect(objRect) {
		this._objRect = {};
		Object.extend(this._objRect, objRect);
	},

	hasRole : function(role) {
		var roles = this.roles;
		for (var i = 0; i < arguments.length; ++i)
			if (roles.indexOf(arguments[i]) != -1) return true;
	},

	getProperty : function(propName, callback) {
		try {
			return this.props[propName].call(this, callback);
		}
		catch(ex) {
			Log.exception(ex, "In TestObject.getProperty, propName: " + propName);
		}
	},

	execute : function(cmd, args, onScrollIntoView) {
		var retCode = TestObject.RETCODE_FAILED;

		if (!this.object && this.testObjId != TestObject.ID_BROWSER)
			return this.endExecute(TestObject.RETCODE_FAILED);

		if (typeof this.cmds[cmd] != "function")
			return this.endExecute(TestObject.RETCODE_NOT_IMPLEMENTED, { cmd : cmd });

		args = args || {};
		if (this.cmds[cmd].schema) {
			var schema = typeof this.cmds[cmd].schema == "function" ? this.cmds[cmd].schema(this) : this.cmds[cmd].schema;
			var errors = Schema.validate(args, schema);
			if (errors.length)
				return this.endExecute(TestObject.RETCODE_INVALID_ARGUMENTS, { cmd : cmd, arguments : errors });
		}
		// Cmds are responsible for calling endExecute()
		try {
			if (this.cmds["Scroll into View"])
				Element.scrollIntoView(this.object);
			if (onScrollIntoView)
				onScrollIntoView();
			var ETA = this.cmds[cmd].call(this, args);
			return typeof ETA == "number" ? ETA : 0;
		} catch(ex) {
			Log.exception(ex, "Catch exception in TestObject cmd execution");
			return this.endExecute(TestObject.RETCODE_FAILED);
		}
	},

	prepareToBeSerializedResistanceIsFutile : function(){
		var str = "";
		for (var prop in this) {
			if (TestObject.isOwnProperty(this,prop) && !activeRepo.hasOwnProperty(prop) && typeof this[prop] == "function") {
				str = str.concat(str ? ",\n" : "", prop, ":", this[prop].toString());
			}
		}
		str += '\n';

		var methodsElem = find("xpath:methods", this)[0];
		if (str) {
			if (methodsElem) {
				while (methodsElem.hasChildNodes())
					methodsElem.removeChild(methodsElem.firstChild);
			}
			else
				methodsElem = this.appendChild(this.ownerDocument.createElement("methods"));
			methodsElem.appendChild(methodsElem.ownerDocument.createCDATASection("{" + str + "}"));
		}
		else if (methodsElem)
			this.removeChild(methodsElem);

		var idents = find("xpath:idents/ident", this);
		for (var i = 0; i < idents.length; i++)
			idents[i].serialize();	// Serialized data is stored in the testObject itself
	},

	evaluate : function(context, maxProfileId, autoUpdate) {
		if (this.testObjId == TestObject.ID_BROWSER)
			return [ activeBrowser ];
		return this.ident.evaluate(context || Window.getActiveTab(), maxProfileId, autoUpdate);
	},

	addRole : function(role, beforeRole) {
		var rolesElem = find("xpath:roles", this)[0] || this.appendChild(TestObject._toDoc.createElement("roles"));
		var roleElem = find('xpath:roles/role[text()="' + role + '"]', this)[0];
		var beforeRoleElem = (beforeRole && find('xpath:roles/role[text()="' + beforeRole + '"]', this)[0]) || null;

		if (roleElem) rolesElem.insertBefore(roleElem, beforeRoleElem);
		else {
			var newRoleElem = TestObject._toDoc.createElement("role");
			newRoleElem.textContent = role;
			rolesElem.insertBefore(newRoleElem, beforeRoleElem);
			Object.extend(this.cmds, Roles.Cmds[role]);
			Object.extend(this.props, Roles.Props[role]);
		}

		if (roleElem || beforeRoleElem) {
			this.cmds = {};
			var roles = this.roles;
			for (var i = 0; i < roles.length; i++) {
				if (Roles.Cmds[roles[i]]) Object.extend(this.cmds, Roles.Cmds[roles[i]]);
				if (Roles.Props[roles[i]]) Object.extend(this.props, Roles.Props[roles[i]]);
			}
		}
	},

	endExecute : function(retCode, props) {
		if (!(retCode >= 0 && retCode <= 5))
			retCode = TestObject.RETCODE_FAILED;
		else if (retCode == TestObject.RETCODE_INVALID_ARGUMENTS_RUNTIME)
			props = { arguments : [ new Schema.Error("Runtime Invalid Value", props.arg, props.values, props.args[props.arg]) ]};
		props = props || {};
		props.retCode = retCode;
		Event.dispatch("testObject/endExecute", this, { msgArgs: props }, Event.DISPATCH_SYNC);
		return retCode;
	},

	delegateExecute : function(obj, cmd, args) {
		var testObj = TestObject(obj, TestObject.TO_CREATE), $this = this;
		testObj.addRole("element");
		Query(testObj).addOnetimeEventListener("testObject/endExecute",
			function(e) {
				$this.endExecute(e.msgArgs.retCode == TestObject.RETCODE_SUCCEEDED ?
					e.msgArgs.retCode : TestObject.RETCODE_FAILED); });
		testObj.execute(cmd, args);
	},

	abort : function() { Event.dispatch("testObject/abort", this); },

	generateName : function() {
		function checkName(str) {
			if (!str) return;
			str = String(str);
			str = str.replace(/<.*>/, "");				// Remove markup
			str = str.normalizeSpace();
			str = str.replace(/_|-/g, ' ');				// Convert '-'/'_' to whitespaces
			str = str.replace(/\([0-9]+\)/g, "");		// Remove parentheses with only numbers inside
			str = str.trim();
			str = str.replace(/:$/, "");				// Remove trailing colon

			if (str.length <= 0 && str.length > 150) return;
			if (str.length > 25) {
				var space = str.indexOf(" ", 20);
				str = str.substr(0, space != -1 ? space : 25) + "...";
			}
			return str;
		}

		if (!this.icon && this.object) {
			var width = this.object.offsetWidth, height = this.object.offsetHeight;
			if (width > 6 && height > 6 && height < 41 && width / height > 0.8 &&
				width / height < 1.2 && !Element.getVisibleText(this.object).trim())
					this.generateIcon();
		}

		if (this.name || !this.object) return this.name;
		var nameGens = RuleEngine.getRules("testObject/name", this.object, this);

		for (var i = 0; i < nameGens.length; i++) {
			if (nameGens[i].disabled) continue;
			var str = nameGens[i].name.evaluate(this.object, this)[0];
			nameGens[i].notifyExec(this.object, str);
			//FF || IE
			if (str instanceof Element || (str && str.document && str instanceof Document.parentWindow(str.document).Element)) {
				if (str == this.object) continue;
				var testObj = TestObject(str, TestObject.TO_NO_PERSISTENCE);
				if (!testObj.name) testObj.generateName();
				str = testObj.name;
			}
			else str = checkName(str);

			if (str) {
				this.setAttribute("autoName", str);
				return this.name;
			}
		}

		if (!this.icon && width > 6 && height > 6 && height < 51 && width / height > 0.7 &&  width / height < 2.5)
			this.generateIcon();

		var objs = find('xpath://testObject[@fallbackName and roles/role[last()]="' + lastItem(this.roles) + '"]', activeRepo);
		this.setAttribute("fallbackName", lastItem(this.roles).concat(" (", objs.length + 1, ")"));
		return this.name;
	},

	generateIcon : function() {
		TestObject._generateIcon(this, canvas);
	},

	toString : function() {	return "".concat('[TestObject "', this.name, '"]');	},

	toJSON : function() { return this.testObjId; }
}

function initTestObjectFromElem(testObjElem) {
	var testObj = TestObject._toDoc.importNode(testObjElem, true);
	testObj.addEventListener("DOMAttrModified", TestObject.onTOAttrModified, false);
	Object.setObjProto(testObj, TestObject.prototype);

	if (activeRepo[testObj.testObjId]) return;

	testObj.cmds = {};
	testObj.props = {};

	var roles = testObj.roles;
	for (var j = 0; j < roles.length; j++) {
		if (Roles.Cmds[roles[j]])
			Object.extend(testObj.cmds, Roles.Cmds[roles[j]]);
		if (Roles.Props[roles[j]])
			Object.extend(testObj.props, Roles.Props[roles[j]]);
	}

	var methods = find("xpath:methods", testObj)[0];
	if (methods) {
		try { Object.extend(testObj, window.eval("(" + methods.textContent + ")"));
		} catch(ex) { Log.exception(ex); }
	}

	activeRepo.appendChild(testObj);
	activeRepo[testObj.testObjId] = testObj;
	return testObj;
}

function initTestObjectIdents(testObj) {
	var idents = find("xpath:idents/ident", testObj);
	for (var i = 0; i < idents.length; i++)
		Ident(testObj, idents[i]);
}

Query.statics.testObj = function(query) { return TestObject("testObj:" + query); };

Query(function(rule) { if (rule.type == "Role Generators") return rule; }).addEventListener(
	"registry/rule/added", function() {
		this.selector = Query(this.selector, { rule : this });
		RuleEngine.addRule(this.selector, "testObject/extend", this); });

Query(function(rule) { if (rule.type == "Name Generators") return rule; }).addEventListener(
	"registry/rule/added", function() {
		this.name = Query(this.name);
		RuleEngine.addRule(this.selector, "testObject/name", this); });

var canvas = document.createElementNS("http://www.w3.org/1999/xhtml", "canvas");

TestObject._toDoc = document.implementation.createDocument(null, null, null);
TestObject._toDoc.isXml = true;

TestObject._toDoc.appendChild(TestObject._toDoc.createElement("testObjectsContainer"));

// Give the TestObjects the properties of XmlElement and of the TestObject (not relevant to IE (no __proto__))
if (TestObject.prototype.__proto__)
	TestObject.prototype.__proto__ = TestObject._toDoc.createElement("YouKilledMyFatherPrepareToDie").__proto__;

window.LoadTestObjectWrapper();
var DEFAULT_REPOSITORY = "__default__";
var id2rep = {}, activeRepo;

var activeNameOfTOPtr, loadingTiemstamp = (new Date()).getTime();
TestObject.switchRepo(DEFAULT_REPOSITORY);

Query.addEventListener("RRE/loaded repository/clear", function() {
	activeRepo[TestObject.ID_BROWSER] = TestObject(TestObject.getActiveBrowserObj(), TestObject.TO_CREATE);
	activeRepo[TestObject.ID_BROWSER].setAttribute("testObjId", TestObject.ID_BROWSER);
	activeRepo[TestObject.ID_BROWSER].setAttribute("manualName", "browser");
	activeRepo[TestObject.ID_BROWSER].addRole("browser");
}, true);

Query.addEventListener("TC.endIteration", function() {
	if (!RTS("RTS.TruClient.SimulateNewUser"))
		return;

	for (var testObj in activeRepo) {
		if (testObj.tagName == "testObject")
			testObj.object = null;
	}
});

Query.addEventListener("loader/exposeRRE", function(e) {
	e.scope.TestObject = TestObject;
});

var serializer = new XMLSerializer();

})();
