// (c) Copyright 2012 Hewlett-Packard Development Company, L.P.

(function() {

"use strict";

window.TestObject = function(obj, flags) {
	// Is obj a TestObject instance?
	if (!obj || obj.tagName == "testObject")
		return obj;

	// If obj is a TestObject testObjId, look it up in the repository
	if (typeof obj == "string")
		return activeRepo[obj];

	// Is the obj the browser?
	if (obj == TestObject.getActiveBrowserObj() && activeRepo[TestObject.ID_BROWSER])
		return activeRepo[TestObject.ID_BROWSER];

	// At this point obj is an AUT DOM element
	flags = flags || TestObject.TO_DEFAULT;

	// Check if obj has a valid TestObject link attached to it that still matches the object
	// Note : bitmap operation
	if (flags & TestObject.TO_CHECK_LINK && obj[TestObject.nameOfTOPtr] && obj[TestObject.nameOfTOPtr].tagName == "testObject") {
		var testObj = obj[TestObject.nameOfTOPtr];

		if (testObj == activeRepo[testObj.testObjId] && testObj.ident.checkTestObj(obj)) {
			testObj.object = obj;
			return testObj;
		}
	}

	// Search the repository for a matching TestObject
	// Note : bitmap operation
	if (flags & TestObject.TO_SEARCH_REPOSITORY) {
		for (var testObj = activeRepo.firstChild; testObj; testObj = testObj.nextSibling) {
			if (testObj.ident.isMatch(obj)) {
				testObj.object = obj;
				return testObj;
			}
		}
	}

	return null;
}

Object.extend(TestObject, {
	TO_DEFAULT							: 0xFFFF,
	TO_CREATE							: 1 << 0,
	TO_CREATE_IDENT						: 1 << 1,
	TO_EXTEND							: 1 << 2,
	TO_CHECK_LINK						: 1 << 3,
	TO_SEARCH_REPOSITORY				: 1 << 4,
	TO_ADD_TO_REPOSITORY				: 1 << 5,

	ID_BROWSER							: "testObj:{00000000-0000-0000-0000-000000000001}",

	RETCODE_SUCCEEDED					: 0,
	RETCODE_FAILED						: 1,
	RETCODE_NOT_IMPLEMENTED				: 2,
	RETCODE_INVALID_ARGUMENTS			: 3,
	RETCODE_INVALID_ARGUMENTS_RUNTIME	: 4,
	RETCODE_ABORTED						: 5
});

TestObject.TO_NO_PERSISTENCE = (TestObject.TO_CREATE | TestObject.TO_EXTEND | TestObject.TO_CHECK_LINK | TestObject.TO_SEARCH_REPOSITORY);

Object.extend(TestObject, {
	import : function(testObjsElem) {

		var xmlDoc = Utils.parseXml(testObjsElem);

		var newTestObjsElems = find("xpath:testObject", xmlDoc.documentElement);
		var newTestObjs = [];
		for (var i = 0; i < newTestObjsElems.length; i++) {
			var newTestObj = initTestObjectFromElem(newTestObjsElems[i]);
			if (newTestObj) newTestObjs.push(newTestObj);
		}
		// Creating identificaion in second pass as they can reference other test objects
		for (var i = 0; i < newTestObjs.length; ++i){
			initTestObjectIdents(newTestObjs[i]);
		}
	},

	export : function(rootTOs) {
		if (!rootTOs) {
			var testObjs = activeRepo;
			for (var testObj = testObjs.firstChild; testObj; testObj = testObj.nextSibling)
				testObj.prepareToBeSerializedResistanceIsFutile();
		}
		else {
			var testObjs = TestObject._toDoc.createElement("testObjects");
			rootTOs.forEach(function(testObj) {
				testObj.prepareToBeSerializedResistanceIsFutile();
				testObjs.appendChild(testObj.cloneNode(true)); });
		}

		return serializer.serializeToString(testObjs);
	},

	activate : function(testObjId) {
		Log.debug("WebTestObjectStub.js, activate testObjId: " + testObjId);
		TestObject.activeTO = activeRepo[testObjId];
		return !! TestObject.activeTO;
	},

	//TODO message that TO changed
	onTOAttrModified : function (e) {
		if (e.target.tagName == "testObject")
			Event.dispatch("testObject/attrModified", e.target, { origEvent : e });
	},

	initRepo : function() {
		var newTestObjs = TestObject._toDoc.createElement("testObjects");
		activeRepo = TestObject._toDoc.firstChild.appendChild(newTestObjs);
	},

	get nameOfTOPtr()	{ return activeNameOfTOPtr; },

	toString : function() { return "The TestObject class"}
});

TestObject.prototype = {
	get object() {
		TestObject.checkObjectReachability(this);
		return this._object;
	},

	set object(obj) {
		this._object = obj;
		if (obj && TestObject.isWrappedJSObject(obj)) obj[TestObject.nameOfTOPtr] = this;
	},

	get name() {
		return this.getAttribute("manualName") || this.getAttribute("autoName") || this.getAttribute("fallbackName") || "";
	},

	set name(value)			{ this.setAttribute("manualName", value); },

	get testObjId()				{ return this.getAttribute("testObjId"); },

	get roles() {
		var retVal = [], roles = find("xpath:roles/role", this);
		for (var i = 0; i < roles.length; i++)
			retVal.push(roles[i].textContent);
		return retVal;
	},

	get refreshedObject() {
		var obj = this.object;	// object getter ensures that parent window of obj != null
		if (obj && TestObject.isObjExist(obj) &&
			!(obj.ownerDocument.compareDocumentPosition(obj) & Node.DOCUMENT_POSITION_DISCONNECTED) &&
				this.ident.checkTestObj(obj, Ident.PROFILE_EXACT)) return obj;

		var objs = this.evaluate();
		this.object = objs.length == 1 ? objs[0] : null;
		return this.object;
	},

	get ident()				{ return find("xpath:idents/ident[@type=../@active]", this)[0];	},
	set ident(val)			{ find("xpath:idents", this)[0].setAttribute("active", val.getAttribute("type")); },

	hasRole : function(role) {
		var roles = this.roles;
		for (var i = 0; i < arguments.length; ++i)
			if (roles.indexOf(arguments[i]) != -1) return true;
	},

	getProperty : function(propName, callback) {
		try {
			return this.props[propName].call(this, callback);
		}
		catch(ex) {
			Log.exception(ex, "Catch exception in TestObject getProperty " + propName);
		}
	},

	execute : function(cmd, args, onScrollIntoView) {
		Log.debug("start execute on TO");

		if (!this.object && this.testObjId != TestObject.ID_BROWSER) {
			Log.debug("TO execution failed");
			return this.endExecute(TestObject.RETCODE_FAILED, { extRetInfo: "Object Not Found", objectsInfo: {length: 0} });
		}
		if (typeof this.cmds[cmd] != "function") {
			Log.debug("TO execution not implemented");
			return this.endExecute(TestObject.RETCODE_NOT_IMPLEMENTED, { cmd : cmd });
		}
		args = args || {};
		if (this.cmds[cmd].schema) {
			Log.debug("TO execution: validating schema");
			var schema = typeof this.cmds[cmd].schema == "function" ? this.cmds[cmd].schema(this) : this.cmds[cmd].schema;
			var errors = Schema.validate(args, schema);
			if (errors.length) {
				Log.debug("TO execution: schema errors")
				return this.endExecute(TestObject.RETCODE_INVALID_ARGUMENTS, { cmd : cmd, arguments : errors });
			}
		}
		// Cmds are responsible for calling endExecute()
		try {
			Log.debug("TO execution: calling command");
			if (this.cmds["Scroll into View"])
				Element.scrollIntoView(this.object);
			if (onScrollIntoView)
				onScrollIntoView();
			var ETA = this.cmds[cmd].call(this, args);
			return typeof ETA == "number" ? ETA : 0;
		} catch(ex) {
			Log.debug("TO execution: exception " + ex.toString());
			Log.exception(ex, "Catch exception in TestObject cmd");
			return this.endExecute(TestObject.RETCODE_FAILED);
		}
	},

	prepareToBeSerializedResistanceIsFutile : function(){
		var str = "";
		for (var prop in this)
			if (TestObject.isOwnProperty(this,prop) && !activeRepo.hasOwnProperty(prop) && typeof this[prop] == "function")
				str = str.concat(str ? ",\n" : "", prop, ":", this[prop].toString());
		str += '\n';

		var methodsElem = find("xpath:methods", this)[0];
		if (str) {
			if (methodsElem)
				while (methodsElem.hasChildNodes()) methodsElem.removeChild(methodsElem.firstChild);
			else
				methodsElem = this.appendChild(this.ownerDocument.createElement("methods"));
			methodsElem.appendChild(methodsElem.ownerDocument.createCDATASection("{" + str + "}"));
		}
		else if (methodsElem)
			this.removeChild(methodsElem);

		var idents = find("xpath:idents/ident", this);
		for (var i = 0; i < idents.length; i++) idents[i].serialize();	// Serialized data is stored in the testObject itself
	},

	evaluate : function(context, maxProfileId, autoUpdate) {
		if (this.testObjId == TestObject.ID_BROWSER)
			return [ activeBrowser ];
		return this.ident.evaluate(context || Window.getActiveTab(), maxProfileId, autoUpdate);
	},

	addRole : function(role, beforeRole) {
		var rolesElem = find("xpath:roles", this)[0] || this.appendChild(TestObject._toDoc.createElement("roles"));
		var roleElem = find('xpath:roles/role[text()="' + role + '"]', this)[0];
		var beforeRoleElem = (beforeRole && find('xpath:roles/role[text()="' + beforeRole + '"]', this)[0]) || null;

		if (roleElem) rolesElem.insertBefore(roleElem, beforeRoleElem);
		else {
			var newRoleElem = TestObject._toDoc.createElement("role");
			newRoleElem.textContent = role;
			rolesElem.insertBefore(newRoleElem, beforeRoleElem);
			Object.extend(this.cmds, Roles.Cmds[role]);
			Object.extend(this.props, Roles.Props[role]);
		}

		if (roleElem || beforeRoleElem) {
			this.cmds = {};
			var roles = this.roles;
			for (var i = 0; i < roles.length; i++) {
				if (Roles.Cmds[roles[i]]) Object.extend(this.cmds, Roles.Cmds[roles[i]]);
				if (Roles.Props[roles[i]]) Object.extend(this.props, Roles.Props[roles[i]]);
			}
		}
	},

	endExecute : function(retCode, props) {
		if (!(retCode >= 0 && retCode <= 5))
			retCode = TestObject.RETCODE_FAILED;
		else if (retCode == TestObject.RETCODE_INVALID_ARGUMENTS_RUNTIME)
			props = { arguments : [ new Schema.Error("Runtime Invalid Value", props.arg, props.values, props.args[props.arg]) ]};
		props = props || {};
		props.retCode = retCode;
		DRRECommunication.send("testObject/endExecute", this, { msgArgs : props });
		ArgsContext.setProp("object", undefined);
		ArgsContext.setProp("FuncArgs", undefined);
		return retCode;
	},

/*	delegateExecute : function(obj, cmd, args) {
		var testObj = TestObject(obj, TestObject.TO_CREATE), $this = this;
		testObj.addRole("element");
		Query(testObj).addOnetimeEventListener("testObject/endExecute",
			function(e) {
				$this.endExecute(e.retCode == TestObject.RETCODE_SUCCEEDED ? e.retCode : TestObject.RETCODE_FAILED); });
		testObj.execute(cmd, args);
	},*/

	//TODO: need to be sent from background
	abort : function() { Event.dispatch("testObject/abort", this); },

	toString : function() {	return "".concat('[TestObject "', this.name, '"]');	},

	toJSON : function() { return this.testObjId; }
}

function initTestObjectFromElem(testObjElem) {
	Log.debug("initTestObjectFromElem testObjElem = " + testObjElem);
	var testObj = TestObject._toDoc.importNode(testObjElem, true);
	Log.debug("initTestObjectFromElem testObj = " + testObj);
	testObj.addEventListener("DOMAttrModified", TestObject.onTOAttrModified, false);
	Object.setObjProto(testObj, TestObject.prototype);

	if (activeRepo[testObj.testObjId])
		return testObj;

	testObj.cmds = {};
	testObj.props = {};

	var roles = testObj.roles;
	for (var j = 0; j < roles.length; j++) {
		if (Roles.Cmds[roles[j]]) Object.extend(testObj.cmds, Roles.Cmds[roles[j]]);
		if (Roles.Props[roles[j]]) Object.extend(testObj.props, Roles.Props[roles[j]]);
	}

	var methods = find("xpath:methods", testObj)[0];
	if (methods) {
		try { Object.extend(testObj, window.eval("(" + methods.textContent + ")"));
		} catch(ex) { Log.exception(ex); }
	}

	activeRepo.appendChild(testObj);
	activeRepo[testObj.testObjId] = testObj;
	return testObj;
}

function initTestObjectIdents(testObj) {
	var idents = find("xpath:idents/ident", testObj);
	Log.debug("initTestObjectIdents: idents = " +  idents);
	for (var i = 0; i < idents.length; i++)
		Ident(testObj, idents[i]);
}

Query.statics.testObj = function(query) { return TestObject("testObj:" + query); };


TestObject._toDoc = document.implementation.createDocument(null, null, null);
TestObject._toDoc.isXml = true;


TestObject._toDoc.appendChild(TestObject._toDoc.createElement("testObjectsContainer"));
//if not IE
if (TestObject.prototype.__proto__)
	TestObject.prototype.__proto__ = TestObject._toDoc.createElement("testObject").__proto__;
window.LoadTestObjectWrapper();

var  activeRepo;
var loadingTiemstamp = (new Date()).getTime();
var activeNameOfTOPtr = "HP TruClient " + loadingTiemstamp;
TestObject.initRepo();
/*
Query.addEventListener("RRE/loaded", function() {
	activeRepo[TestObject.ID_BROWSER] = TestObject(TestObject.getActiveBrowserObj(), TestObject.TO_CREATE);
	activeRepo[TestObject.ID_BROWSER].setAttribute("testObjId", TestObject.ID_BROWSER);
	activeRepo[TestObject.ID_BROWSER].setAttribute("manualName", "browser");
	activeRepo[TestObject.ID_BROWSER].addRole("browser");
}, true);
*/
var serializer = new XMLSerializer();

})();
